@@ -19,15 +19,17 @@
 
 from __future__ import annotations
 
-from typing import List, TYPE_CHECKING
+from typing import List, TYPE_CHECKING, Union
 
 import fenics
 import numpy as np
 import ufl
+import ufl.algorithms
 
 from cashocs import _exceptions
 from cashocs import _utils
 from cashocs._forms import form_handler
+from cashocs._optimization import cost_functional
 
 if TYPE_CHECKING:
     from cashocs._optimization import optimal_control
@@ -42,6 +44,23 @@ class ControlFormHandler(form_handler.FormHandler):
      the optimization (solution) algorithms.
     """
 
+    idx_active: List
+    idx_active_lower: List
+    idx_active_upper: List
+    idx_inactive: List
+    adjoint_sensitivity_eqs_picard: List[ufl.Form]
+    adjoint_sensitivity_eqs_rhs: List[ufl.Form]
+    w_1: List[ufl.Form]
+    w_2: List[ufl.Form]
+    w_3: List[ufl.Form]
+    hessian_rhs: List[ufl.Form]
+    states_prime: List[fenics.Function]
+    adjoints_prime: List[fenics.Function]
+    test_directions: List[fenics.Function]
+    trial_functions_control: List[fenics.Function]
+    test_functions_control: List[fenics.Function]
+    temp: List[fenics.Function]
+
     def __init__(
         self, optimization_problem: optimal_control.OptimalControlProblem
     ) -> None:
@@ -56,40 +75,33 @@ class ControlFormHandler(form_handler.FormHandler):
         # Initialize the attributes from the arguments
         self.controls = optimization_problem.controls
         self.riesz_scalar_products = optimization_problem.riesz_scalar_products
+        self.control_bcs_list = optimization_problem.control_bcs_list
         self.control_constraints = optimization_problem.control_constraints
         self.require_control_constraints = (
             optimization_problem.require_control_constraints
         )
 
-        self.idx_active = None
-        self.idx_active_lower = None
-        self.idx_active_upper = None
-        self.idx_inactive = None
-
-        self.w_1 = None
-        self.w_2 = None
-        self.w_3 = None
-        self.hessian_rhs = None
-
         self.control_dim = len(self.controls)
         self.control_spaces = [x.function_space() for x in self.controls]
 
-        self.gradient = [
-            fenics.Function(function_space) for function_space in self.control_spaces
-        ]
+        self.gradient = _utils.create_function_list(self.control_spaces)
 
-        # Define the necessary functions
-        self.states_prime = [
-            fenics.Function(function_space) for function_space in self.state_spaces
-        ]
-        self.adjoints_prime = [
-            fenics.Function(function_space) for function_space in self.adjoint_spaces
-        ]
+        self._init_helpers()
+        self._compute_gradient_equations()
 
-        self.test_directions = [
-            fenics.Function(function_space) for function_space in self.control_spaces
-        ]
+        if self.opt_algo.casefold() == "newton":
+            self.compute_newton_forms()
 
+        self.setup_assemblers(
+            self.riesz_scalar_products, self.gradient_forms_rhs, self.control_bcs_list
+        )
+
+    def _init_helpers(self) -> None:
+        """Initializes the helper functions needed for the form handler."""
+        self.states_prime = _utils.create_function_list(self.state_spaces)
+        self.adjoints_prime = _utils.create_function_list(self.adjoint_spaces)
+        self.test_directions = _utils.create_function_list(self.control_spaces)
+        self.temp = _utils.create_function_list(self.control_spaces)
         self.trial_functions_control = [
             fenics.TrialFunction(function_space)
             for function_space in self.control_spaces
@@ -99,28 +111,61 @@ class ControlFormHandler(form_handler.FormHandler):
             for function_space in self.control_spaces
         ]
 
-        self.temp = [
-            fenics.Function(function_space) for function_space in self.control_spaces
-        ]
+    def setup_assemblers(
+        self,
+        scalar_product_forms: List[ufl.Form],
+        derivatives: List[ufl.Form],
+        bcs: Union[List[List[fenics.DirichletBC]], List[None]],
+    ) -> None:
+        """Sets up the assemblers and matrices for the projection of the gradient.
 
-        # Compute the necessary equations
-        self._compute_gradient_equations()
+        Args:
+            scalar_product_forms: The weak form of the scalar product.
+            derivatives: The weak form of the derivative.
+            bcs: The boundary conditions for the projection.
 
-        if self.opt_algo.casefold() == "newton":
-            self.compute_newton_forms()
+        """
+        modified_scalar_product_forms = _utils.bilinear_boundary_form_modification(
+            scalar_product_forms
+        )
+        try:
+            self.assemblers = []
+            for i in range(self.control_dim):
+                assembler = fenics.SystemAssembler(
+                    modified_scalar_product_forms[i],
+                    derivatives[i],
+                    bcs[i],
+                )
+                assembler.keep_diagonal = True
+                self.assemblers.append(assembler)
+        except (AssertionError, ValueError):
+            self.assemblers = []
+            for i in range(self.control_dim):
+                estimated_degree = np.maximum(
+                    ufl.algorithms.estimate_total_polynomial_degree(
+                        self.riesz_scalar_products[i]
+                    ),
+                    ufl.algorithms.estimate_total_polynomial_degree(
+                        self.gradient_forms_rhs[i]
+                    ),
+                )
+                assembler = fenics.SystemAssembler(
+                    modified_scalar_product_forms[i],
+                    derivatives[i],
+                    bcs[i],
+                    form_compiler_parameters={"quadrature_degree": estimated_degree},
+                )
+                assembler.keep_diagonal = True
+                self.assemblers.append(assembler)
 
-        # Initialize the scalar products
         fenics_scalar_product_matrices = [
             fenics.PETScMatrix() for _ in range(self.control_dim)
         ]
 
         for i in range(self.control_dim):
-            fenics.assemble(
-                self.riesz_scalar_products[i],
-                keep_diagonal=True,
-                tensor=fenics_scalar_product_matrices[i],
-            )
+            self.assemblers[i].assemble(fenics_scalar_product_matrices[i])
             fenics_scalar_product_matrices[i].ident_zeros()
+
         self.riesz_projection_matrices = [
             fenics_scalar_product_matrices[i].mat() for i in range(self.control_dim)
         ]
@@ -194,6 +239,16 @@ class ControlFormHandler(form_handler.FormHandler):
                         )
                     )[0]
                 )
+                self.idx_inactive.append(
+                    np.nonzero(
+                        np.logical_and(
+                            self.controls[j].vector()[:]
+                            > self.control_constraints[j][0].vector()[:],
+                            self.controls[j].vector()[:]
+                            < self.control_constraints[j][1].vector()[:],
+                        )
+                    )[0]
+                )
             else:
                 self.idx_active_lower.append([])
                 self.idx_active_upper.append([])
@@ -203,11 +258,6 @@ class ControlFormHandler(form_handler.FormHandler):
             )
             temp_active.sort()
             self.idx_active.append(temp_active)
-            self.idx_inactive.append(
-                np.setdiff1d(
-                    np.arange(self.control_spaces[j].dim()), self.idx_active[j]
-                )
-            )
 
     def restrict_to_active_set(
         self, a: List[fenics.Function], b: List[fenics.Function]
@@ -229,13 +279,17 @@ class ControlFormHandler(form_handler.FormHandler):
         for j in range(self.control_dim):
             if self.require_control_constraints[j]:
                 self.temp[j].vector().vec().set(0.0)
+                self.temp[j].vector().apply("")
                 self.temp[j].vector()[self.idx_active[j]] = a[j].vector()[
                     self.idx_active[j]
                 ]
+                self.temp[j].vector().apply("")
                 b[j].vector().vec().aypx(0.0, self.temp[j].vector().vec())
+                b[j].vector().apply("")
 
             else:
                 b[j].vector().vec().set(0.0)
+                b[j].vector().apply("")
 
         return b
 
@@ -259,14 +313,18 @@ class ControlFormHandler(form_handler.FormHandler):
         for j in range(self.control_dim):
             if self.require_control_constraints[j]:
                 self.temp[j].vector().vec().set(0.0)
+                self.temp[j].vector().apply("")
                 self.temp[j].vector()[self.idx_inactive[j]] = a[j].vector()[
                     self.idx_inactive[j]
                 ]
+                self.temp[j].vector().apply("")
                 b[j].vector().vec().aypx(0.0, self.temp[j].vector().vec())
+                b[j].vector().apply("")
 
             else:
                 if not b[j].vector().vec().equal(a[j].vector().vec()):
                     b[j].vector().vec().aypx(0.0, a[j].vector().vec())
+                    b[j].vector().apply("")
 
         return b
 
@@ -291,68 +349,21 @@ class ControlFormHandler(form_handler.FormHandler):
                 a[j].vector().vec().pointwiseMin(
                     self.control_constraints[j][1].vector().vec(), a[j].vector().vec()
                 )
+                a[j].vector().apply("")
                 a[j].vector().vec().pointwiseMax(
                     a[j].vector().vec(), self.control_constraints[j][0].vector().vec()
                 )
+                a[j].vector().apply("")
 
         return a
 
     def _compute_gradient_equations(self) -> None:
         """Calculates the variational form of the gradient equation."""
         self.gradient_forms_rhs = [
-            fenics.derivative(
-                self.lagrangian_form,
-                self.controls[i],
-                self.test_functions_control[i],
-            )
+            self.lagrangian.derivative(self.controls[i], self.test_functions_control[i])
             for i in range(self.control_dim)
         ]
 
-        if self.use_scalar_tracking:
-            for i in range(self.control_dim):
-                for j in range(self.no_scalar_tracking_terms):
-                    self.gradient_forms_rhs[i] += (
-                        self.scalar_weights[j]
-                        * (
-                            self.scalar_cost_functional_integrand_values[j]
-                            - fenics.Constant(self.scalar_tracking_goals[j])
-                        )
-                        * fenics.derivative(
-                            self.scalar_cost_functional_integrands[j],
-                            self.controls[i],
-                            self.test_functions_control[i],
-                        )
-                    )
-
-        if self.use_min_max_terms:
-            for i in range(self.control_dim):
-                for j in range(self.no_min_max_terms):
-                    if self.min_max_lower_bounds[j] is not None:
-                        term_lower = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                            self.min_max_integrand_values[j]
-                            - self.min_max_lower_bounds[j]
-                        )
-                        self.gradient_forms_rhs[i] += _utils.min_(
-                            fenics.Constant(0.0), term_lower
-                        ) * fenics.derivative(
-                            self.min_max_integrands[j],
-                            self.controls[i],
-                            self.test_functions_control[i],
-                        )
-
-                    if self.min_max_upper_bounds[j] is not None:
-                        term_upper = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                            self.min_max_integrand_values[j]
-                            - self.min_max_upper_bounds[j]
-                        )
-                        self.gradient_forms_rhs[i] += _utils.max_(
-                            fenics.Constant(0.0), term_upper
-                        ) * fenics.derivative(
-                            self.min_max_integrands[j],
-                            self.controls[i],
-                            self.test_functions_control[i],
-                        )
-
     def _compute_sensitivity_equations(self) -> None:
         """Calculates the forms for the (forward) sensitivity equations."""
         # Use replace -> derivative to speed up the computations
@@ -363,7 +374,7 @@ class ControlFormHandler(form_handler.FormHandler):
             for i in range(self.state_dim)
         ]
 
-        self.sensitivity_eqs_lhs = [
+        self.sensitivity_eqs_lhs: List[ufl.Form] = [
             fenics.derivative(
                 self.sensitivity_eqs_temp[i],
                 self.states[i],
@@ -382,7 +393,7 @@ class ControlFormHandler(form_handler.FormHandler):
         # Need to distinguish cases due to empty sum in case state_dim = 1
         if self.state_dim > 1:
             # pylint: disable=invalid-unary-operand-type
-            self.sensitivity_eqs_rhs = [
+            self.sensitivity_eqs_rhs: List[ufl.Form] = [
                 -_utils.summation(
                     [
                         fenics.derivative(
@@ -430,19 +441,11 @@ class ControlFormHandler(form_handler.FormHandler):
     def _compute_first_order_lagrangian_derivatives(self) -> None:
         """Computes the derivative of the Lagrangian w.r.t. the state and control."""
         self.lagrangian_y = [
-            fenics.derivative(
-                self.lagrangian_form,
-                self.states[i],
-                self.test_functions_state[i],
-            )
+            self.lagrangian.derivative(self.states[i], self.test_functions_state[i])
             for i in range(self.state_dim)
         ]
         self.lagrangian_u = [
-            fenics.derivative(
-                self.lagrangian_form,
-                self.controls[i],
-                self.test_functions_control[i],
-            )
+            self.lagrangian.derivative(self.controls[i], self.test_functions_control[i])
             for i in range(self.control_dim)
         ]
 
@@ -559,7 +562,16 @@ class ControlFormHandler(form_handler.FormHandler):
 
     def compute_newton_forms(self) -> None:
         """Calculates the needed forms for the truncated Newton method."""
-        if self.use_scalar_tracking or self.use_min_max_terms:
+        if any(
+            isinstance(
+                functional,
+                (
+                    cost_functional.ScalarTrackingFunctional,
+                    cost_functional.MinMaxFunctional,
+                ),
+            )
+            for functional in self.cost_functional_list
+        ):
             raise _exceptions.InputError(
                 "cashocs._forms.ShapeFormHandler",
                 "_compute_newton_forms",
