@@ -20,17 +20,19 @@
 from __future__ import annotations
 
 import abc
-from typing import List, Optional, TYPE_CHECKING, Union
+from typing import Callable, List, TYPE_CHECKING, Union
 
 import fenics
 from petsc4py import PETSc
 import ufl
 
 from cashocs import _utils
+from cashocs._optimization import cost_functional
 
 if TYPE_CHECKING:
-    from cashocs import _optimization as op
-    from cashocs._forms import shape_regularization
+    from cashocs import io
+    from cashocs import types
+    from cashocs._forms import shape_regularization as sr
 
 
 def _get_subdx(
@@ -84,99 +86,45 @@ class FormHandler(abc.ABC):
     # noinspection PyUnresolvedReferences
     riesz_projection_matrices: List[PETSc.Mat]
     uses_custom_scalar_product: bool = False
+    gradient_forms_rhs: List[ufl.Form]
+    bcs_shape: List[fenics.DirichletBC]
+    shape_regularization: sr.ShapeRegularization
+    shape_derivative: ufl.Form
+    scalar_product_matrix: fenics.PETScMatrix
+    mu_lame: fenics.Function
+    config: io.Config
+    scalar_cost_functional_integrands: List[ufl.Form]
+    scalar_cost_functional_integrand_values: List[fenics.Function]
+    states: List[fenics.Function]
+    cost_functional_list: List[types.CostFunctional]
 
-    def __init__(self, optimization_problem: op.OptimizationProblem) -> None:
+    def __init__(self, optimization_problem: types.OptimizationProblem) -> None:
         """Initializes self.
 
         Args:
             optimization_problem: The corresponding optimization problem
 
         """
-        self.bcs_list = optimization_problem.bcs_list
+        self.optimization_problem = optimization_problem
+        self.bcs_list: List[List[fenics.DirichletBC]] = optimization_problem.bcs_list
         self.states = optimization_problem.states
         self.adjoints = optimization_problem.adjoints
         self.config = optimization_problem.config
         self.state_ksp_options = optimization_problem.ksp_options
         self.adjoint_ksp_options = optimization_problem.adjoint_ksp_options
-        self.use_scalar_tracking = optimization_problem.use_scalar_tracking
-        self.use_min_max_terms = optimization_problem.use_min_max_terms
-        self.min_max_forms = optimization_problem.min_max_terms
-        self.scalar_tracking_forms = optimization_problem.scalar_tracking_forms
 
-        self.is_shape_problem = optimization_problem.is_shape_problem
+        self.is_shape_problem: bool = optimization_problem.is_shape_problem
         self.is_control_problem = optimization_problem.is_control_problem
 
-        self.cost_functional_form = optimization_problem.cost_functional_form
+        self.cost_functional_list = optimization_problem.cost_functional_list
         self.state_forms = optimization_problem.state_forms
 
-        self.shape_regularization: Optional[
-            shape_regularization.ShapeRegularization
-        ] = None
-        self.gradient_forms_rhs = None
-        self.shape_derivative = None
-        self.bcs_shape = None
-        self.scalar_product_matrix = None
-        self.mu_lame = None
-
-        self.lagrangian_form = self.cost_functional_form + _utils.summation(
-            self.state_forms
+        self.lagrangian = cost_functional.Lagrangian(
+            self.cost_functional_list, self.state_forms
         )
-        self.cost_functional_shift = 0.0
-
-        if self.use_scalar_tracking:
-            self.scalar_cost_functional_integrands = [
-                d["integrand"] for d in self.scalar_tracking_forms
-            ]
-            self.scalar_tracking_goals = [
-                d["tracking_goal"] for d in self.scalar_tracking_forms
-            ]
-
-            dummy_meshes = [
-                integrand.integrals()[0].ufl_domain().ufl_cargo()
-                for integrand in self.scalar_cost_functional_integrands
-            ]
-            self.scalar_cost_functional_integrand_values = [
-                fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-                for mesh in dummy_meshes
-            ]
-            self.scalar_weights = [
-                fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-                for mesh in dummy_meshes
-            ]
-
-            self.no_scalar_tracking_terms = len(self.scalar_tracking_goals)
-            try:
-                for j in range(self.no_scalar_tracking_terms):
-                    self.scalar_weights[j].vector().vec().set(
-                        self.scalar_tracking_forms[j]["weight"]
-                    )
-            except KeyError:
-                for j in range(self.no_scalar_tracking_terms):
-                    self.scalar_weights[j].vector().vec().set(1.0)
-
-        if self.use_min_max_terms:
-            self.min_max_integrands = [d["integrand"] for d in self.min_max_forms]
-            self.min_max_lower_bounds = [d["lower_bound"] for d in self.min_max_forms]
-            self.min_max_upper_bounds = [d["upper_bound"] for d in self.min_max_forms]
-
-            dummy_meshes = [
-                integrand.integrals()[0].ufl_domain().ufl_cargo()
-                for integrand in self.min_max_integrands
-            ]
-            self.min_max_integrand_values = [
-                fenics.Function(fenics.FunctionSpace(mesh, "R", 0))
-                for mesh in dummy_meshes
-            ]
-
-            self.no_min_max_terms = len(self.min_max_integrands)
-            self.min_max_mu = []
-            self.min_max_lambda = []
-            for j in range(self.no_min_max_terms):
-                self.min_max_mu.append(self.min_max_forms[j]["mu"])
-                self.min_max_lambda.append(self.min_max_forms[j]["lambda"])
+        self.cost_functional_shift: float = 0.0
 
         self.state_dim = len(self.states)
-
         self.state_spaces = [x.function_space() for x in self.states]
         self.adjoint_spaces = [x.function_space() for x in self.adjoints]
 
@@ -186,7 +134,7 @@ class FormHandler(abc.ABC):
         else:
             self.state_adjoint_equal_spaces = False
 
-        self.mesh = self.state_spaces[0].mesh()
+        self.mesh: fenics.Mesh = self.state_spaces[0].mesh()
         self.dx = fenics.Measure("dx", self.mesh)
 
         self.trial_functions_state = [
@@ -209,8 +157,8 @@ class FormHandler(abc.ABC):
         self.state_is_picard = self.config.getboolean("StateSystem", "picard_iteration")
         self.opt_algo = _utils.optimization_algorithm_configuration(self.config)
 
-        self.pre_hook = _hook
-        self.post_hook = _hook
+        self.pre_hook: Callable[..., None] = _hook
+        self.post_hook: Callable[..., None] = _hook
 
         self._compute_state_equations()
         self._compute_adjoint_equations()
@@ -257,7 +205,8 @@ class FormHandler(abc.ABC):
             for i in range(self.state_dim):
                 for j, bc in enumerate(self.bcs_list[i]):
                     idx = bc.function_space().id()
-                    subdx = _get_subdx(self.state_spaces[i], idx, ls=[])
+                    subdx: List[int] = []
+                    _get_subdx(self.state_spaces[i], idx, ls=subdx)
                     adjoint_space = self.adjoint_spaces[i]
                     for num in subdx:
                         adjoint_space = adjoint_space.sub(num)
@@ -281,70 +230,14 @@ class FormHandler(abc.ABC):
                             adjoint_space, bdry_value, bc.sub_domain
                         )
 
-    def _compute_adjoint_scalar_tracking_forms(self) -> None:
-        """Compute the part arising due to scalar_tracking_terms."""
-        if self.use_scalar_tracking:
-            for i in range(self.state_dim):
-                for j in range(self.no_scalar_tracking_terms):
-                    self.adjoint_eq_forms[i] += (
-                        self.scalar_weights[j]
-                        * (
-                            self.scalar_cost_functional_integrand_values[j]
-                            - fenics.Constant(self.scalar_tracking_goals[j])
-                        )
-                        * fenics.derivative(
-                            self.scalar_cost_functional_integrands[j],
-                            self.states[i],
-                            self.test_functions_adjoint[i],
-                        )
-                    )
-
-    def _compute_adjoint_min_max_forms(self) -> None:
-        """Compute the part arising due to min_max_terms."""
-        if self.use_min_max_terms:
-            for i in range(self.state_dim):
-                for j in range(self.no_min_max_terms):
-                    if self.min_max_lower_bounds[j] is not None:
-                        term_lower = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                            self.min_max_integrand_values[j]
-                            - self.min_max_lower_bounds[j]
-                        )
-                        self.adjoint_eq_forms[i] += _utils.min_(
-                            fenics.Constant(0.0), term_lower
-                        ) * fenics.derivative(
-                            self.min_max_integrands[j],
-                            self.states[i],
-                            self.test_functions_adjoint[i],
-                        )
-
-                    if self.min_max_upper_bounds[j] is not None:
-                        term_upper = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                            self.min_max_integrand_values[j]
-                            - self.min_max_upper_bounds[j]
-                        )
-                        self.adjoint_eq_forms[i] += _utils.max_(
-                            fenics.Constant(0.0), term_upper
-                        ) * fenics.derivative(
-                            self.min_max_integrands[j],
-                            self.states[i],
-                            self.test_functions_adjoint[i],
-                        )
-
     def _compute_adjoint_equations(self) -> None:
         """Calculates the weak form of the adjoint equation for use with fenics."""
-        self.adjoint_eq_forms = [
-            fenics.derivative(
-                self.lagrangian_form,
-                self.states[i],
-                self.test_functions_adjoint[i],
-            )
+        self.adjoint_eq_forms: List[ufl.Form] = [
+            self.lagrangian.derivative(self.states[i], self.test_functions_adjoint[i])
             for i in range(self.state_dim)
         ]
 
-        self._compute_adjoint_scalar_tracking_forms()
-        self._compute_adjoint_min_max_forms()
-
-        self.linear_adjoint_eq_forms = [
+        self.linear_adjoint_eq_forms: List[ufl.Form] = [
             ufl.replace(
                 self.adjoint_eq_forms[i],
                 {self.adjoints[i]: self.trial_functions_adjoint[i]},
@@ -393,6 +286,7 @@ class FormHandler(abc.ABC):
         for j in range(len(self.gradient)):
             if not b[j].vector().vec().equal(a[j].vector().vec()):
                 b[j].vector().vec().aypx(0.0, a[j].vector().vec())
+                b[j].vector().apply("")
 
         return b
 
@@ -415,6 +309,7 @@ class FormHandler(abc.ABC):
         """
         for j in range(len(self.gradient)):
             b[j].vector().vec().set(0.0)
+            b[j].vector().apply("")
 
         return b
 
