@@ -20,7 +20,7 @@
 from __future__ import annotations
 
 import itertools
-from typing import List, Optional, TYPE_CHECKING
+from typing import List, Optional, TYPE_CHECKING, Union
 
 import fenics
 import numpy as np
@@ -47,6 +47,9 @@ class ShapeFormHandler(form_handler.FormHandler):
     and the shape derivatives.
     """
 
+    scalar_product_matrix: fenics.PETScMatrix
+    material_derivative: ufl.Form
+
     def __init__(
         self, optimization_problem: shape_optimization.ShapeOptimizationProblem
     ) -> None:
@@ -58,8 +61,10 @@ class ShapeFormHandler(form_handler.FormHandler):
         """
         super().__init__(optimization_problem)
 
-        self.has_cashocs_remesh_flag = optimization_problem.has_cashocs_remesh_flag
-        self.temp_dir = optimization_problem.temp_dir
+        self.has_cashocs_remesh_flag: bool = (
+            optimization_problem.has_cashocs_remesh_flag
+        )
+        self.temp_dir: Optional[str] = optimization_problem.temp_dir
         self.boundaries = optimization_problem.boundaries
         self.shape_scalar_product = optimization_problem.shape_scalar_product
         self.uses_custom_scalar_product = (
@@ -67,8 +72,6 @@ class ShapeFormHandler(form_handler.FormHandler):
         )
         deformation_space = optimization_problem.deformation_space
 
-        self.scalar_product_matrix: Optional[fenics.PETScMatrix] = None
-
         self.degree_estimation = self.config.getboolean(
             "ShapeGradient", "degree_estimation"
         )
@@ -90,7 +93,9 @@ class ShapeFormHandler(form_handler.FormHandler):
         self.gradient = [fenics.Function(self.deformation_space)]
         self.test_vector_field = fenics.TestFunction(self.deformation_space)
 
-        self.shape_regularization = shape_regularization.ShapeRegularization(self)
+        self.shape_regularization: shape_regularization.ShapeRegularization = (
+            shape_regularization.ShapeRegularization(self)
+        )
 
         fixed_dimensions = self.config.getlist("ShapeGradient", "fixed_dimensions")
         self.use_fixed_dimensions = False
@@ -107,30 +112,9 @@ class ShapeFormHandler(form_handler.FormHandler):
         self._compute_shape_gradient_forms()
         self._setup_mu_computation()
 
-        retry_assembler_setup = False
-        if not self.degree_estimation:
-            try:
-                self.assembler = fenics.SystemAssembler(
-                    self.riesz_scalar_product, self.shape_derivative, self.bcs_shape
-                )
-            except (AssertionError, ValueError):
-                retry_assembler_setup = True
-
-        if retry_assembler_setup or self.degree_estimation:
-            self.estimated_degree = np.maximum(
-                ufl.algorithms.estimate_total_polynomial_degree(
-                    self.riesz_scalar_product
-                ),
-                ufl.algorithms.estimate_total_polynomial_degree(self.shape_derivative),
-            )
-            self.assembler = fenics.SystemAssembler(
-                self.riesz_scalar_product,
-                self.shape_derivative,
-                self.bcs_shape,
-                form_compiler_parameters={"quadrature_degree": self.estimated_degree},
-            )
-
-        self.assembler.keep_diagonal = True
+        self.setup_assembler(
+            self.riesz_scalar_product, self.shape_derivative, self.bcs_shape
+        )
         self.fe_scalar_product_matrix = fenics.PETScMatrix()
         self.fe_shape_derivative_vector = fenics.PETScVector()
 
@@ -163,99 +147,46 @@ class ShapeFormHandler(form_handler.FormHandler):
                 "Second order methods are not implemented for shape optimization yet"
             )
 
-    def _compute_scalar_tracking_shape_derivative(self) -> None:
-        """Calculates the shape derivative of scalar_tracking_forms."""
-        if self.use_scalar_tracking:
-            for j in range(self.no_scalar_tracking_terms):
-                self.shape_derivative += fenics.derivative(
-                    self.scalar_weights[j]
-                    * (
-                        self.scalar_cost_functional_integrand_values[j]
-                        - fenics.Constant(self.scalar_tracking_goals[j])
-                    )
-                    * self.scalar_cost_functional_integrands[j],
-                    fenics.SpatialCoordinate(self.mesh),
-                    self.test_vector_field,
-                )
-
-    def _compute_min_max_shape_derivative(self) -> None:
-        """Calculates the shape derivative of min_max_terms."""
-        if self.use_min_max_terms:
-            for j in range(self.no_min_max_terms):
-                if self.min_max_lower_bounds[j] is not None:
-                    term_lower = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                        self.min_max_integrand_values[j] - self.min_max_lower_bounds[j]
-                    )
-                    self.shape_derivative += fenics.derivative(
-                        _utils.min_(fenics.Constant(0.0), term_lower)
-                        * self.min_max_integrands[j],
-                        fenics.SpatialCoordinate(self.mesh),
-                        self.test_vector_field,
-                    )
-
-                if self.min_max_upper_bounds[j] is not None:
-                    term_upper = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                        self.min_max_integrand_values[j] - self.min_max_upper_bounds[j]
-                    )
-                    self.shape_derivative += fenics.derivative(
-                        _utils.max_(fenics.Constant(0.0), term_upper)
-                        * self.min_max_integrands[j],
-                        fenics.SpatialCoordinate(self.mesh),
-                        self.test_vector_field,
-                    )
-
-    # noinspection PyUnresolvedReferences
-    def _add_scalar_tracking_pull_backs(self, coeff: ufl.core.expr.Expr) -> None:
-        """Adds pullbacks for scalar_tracking_forms.
+    def setup_assembler(
+        self,
+        scalar_product: ufl.form,
+        shape_derivative: ufl.form,
+        bcs: Optional[List[fenics.DirichletBC]],
+    ) -> None:
+        """Sets up the assembler for assembling the shape gradient projection.
 
         Args:
-            coeff: The coefficient after which the Lagrangian is differentiated.
+            scalar_product: The weak form of the scalar product
+            shape_derivative: The weak form of the shape derivative
+            bcs: The boundary conditions for the projection
 
         """
-        if self.use_scalar_tracking:
-            for j in range(self.no_scalar_tracking_terms):
-                self.material_derivative += fenics.derivative(
-                    self.scalar_weights[j]
-                    * (
-                        self.scalar_cost_functional_integrand_values[j]
-                        - fenics.Constant(self.scalar_tracking_goals[j])
-                    )
-                    * self.scalar_cost_functional_integrands[j],
-                    coeff,
-                    fenics.dot(fenics.grad(coeff), self.test_vector_field),
+        modified_scalar_product = _utils.bilinear_boundary_form_modification(
+            [scalar_product]
+        )[0]
+        retry_assembler_setup = False
+        if not self.degree_estimation:
+            try:
+                self.assembler = fenics.SystemAssembler(
+                    modified_scalar_product, shape_derivative, bcs
                 )
+            except (AssertionError, ValueError):
+                retry_assembler_setup = True
 
-    # noinspection PyUnresolvedReferences
-    def _add_min_max_pull_backs(self, coeff: ufl.core.expr.Expr) -> None:
-        """Adds pullbacks for min_max_terms.
-
-        Args:
-            coeff: The coefficient after which the Lagrangian is differentiated.
-
-        """
-        if self.use_min_max_terms:
-            for j in range(self.no_min_max_terms):
-                if self.min_max_lower_bounds[j] is not None:
-                    term_lower = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                        self.min_max_integrand_values[j] - self.min_max_lower_bounds[j]
-                    )
-                    self.material_derivative += fenics.derivative(
-                        _utils.min_(fenics.Constant(0.0), term_lower)
-                        * self.min_max_integrands[j],
-                        coeff,
-                        fenics.dot(fenics.grad(coeff), self.test_vector_field),
-                    )
-
-                if self.min_max_upper_bounds[j] is not None:
-                    term_upper = self.min_max_lambda[j] + self.min_max_mu[j] * (
-                        self.min_max_integrand_values[j] - self.min_max_upper_bounds[j]
-                    )
-                    self.material_derivative += fenics.derivative(
-                        _utils.max_(fenics.Constant(0.0), term_upper)
-                        * self.min_max_integrands[j],
-                        coeff,
-                        fenics.dot(fenics.grad(coeff), self.test_vector_field),
-                    )
+        if retry_assembler_setup or self.degree_estimation:
+            estimated_degree = np.maximum(
+                ufl.algorithms.estimate_total_polynomial_degree(
+                    modified_scalar_product
+                ),
+                ufl.algorithms.estimate_total_polynomial_degree(shape_derivative),
+            )
+            self.assembler = fenics.SystemAssembler(
+                modified_scalar_product,
+                shape_derivative,
+                bcs,
+                form_compiler_parameters={"quadrature_degree": estimated_degree},
+            )
+        self.assembler.keep_diagonal = True
 
     # noinspection PyUnresolvedReferences
     def _check_coefficient_id(self, coeff: ufl.core.expr.Expr) -> None:
@@ -273,25 +204,15 @@ class ShapeFormHandler(form_handler.FormHandler):
 
     def _parse_pull_back_coefficients(self) -> None:
         """Parses the coefficients which are available for adding pullbacks."""
-        self.state_adjoint_ids = [coeff.id() for coeff in self.states] + [
+        self.state_adjoint_ids: List[int] = [coeff.id() for coeff in self.states] + [
             coeff.id() for coeff in self.adjoints
         ]
 
-        self.material_derivative_coeffs = []
+        self.material_derivative_coeffs: List[ufl.core.expr.Expr] = []
 
-        for coeff in self.lagrangian_form.coefficients():
+        for coeff in self.lagrangian.coefficients():
             self._check_coefficient_id(coeff)
 
-        if self.use_scalar_tracking:
-            for j in range(self.no_scalar_tracking_terms):
-                for coeff in self.scalar_cost_functional_integrands[j].coefficients():
-                    self._check_coefficient_id(coeff)
-
-        if self.use_min_max_terms:
-            for j in range(self.no_min_max_terms):
-                for coeff in self.min_max_integrands[j].coefficients():
-                    self._check_coefficient_id(coeff)
-
         if len(self.material_derivative_coeffs) > 0:
             _loggers.warning(
                 "Shape derivative might be wrong, if differential operators "
@@ -307,15 +228,10 @@ class ShapeFormHandler(form_handler.FormHandler):
 
             for coeff in self.material_derivative_coeffs:
 
-                self.material_derivative = fenics.derivative(
-                    self.lagrangian_form,
-                    coeff,
-                    fenics.dot(fenics.grad(coeff), self.test_vector_field),
+                self.material_derivative = self.lagrangian.derivative(
+                    coeff, fenics.dot(fenics.grad(coeff), self.test_vector_field)
                 )
 
-                self._add_scalar_tracking_pull_backs(coeff)
-                self._add_min_max_pull_backs(coeff)
-
                 self.material_derivative = ufl.algorithms.expand_derivatives(
                     self.material_derivative
                 )
@@ -330,14 +246,10 @@ class ShapeFormHandler(form_handler.FormHandler):
         A corresponding warning whenever this could be the case is issued.
         """
         # Shape derivative of Lagrangian w/o regularization and pullbacks
-        self.shape_derivative = fenics.derivative(
-            self.lagrangian_form,
-            fenics.SpatialCoordinate(self.mesh),
-            self.test_vector_field,
+        self.shape_derivative = self.lagrangian.derivative(
+            fenics.SpatialCoordinate(self.mesh), self.test_vector_field
         )
 
-        self._compute_scalar_tracking_shape_derivative()
-        self._compute_min_max_shape_derivative()
         self._add_pull_backs()
 
         # Add regularization
@@ -381,8 +293,9 @@ class ShapeFormHandler(form_handler.FormHandler):
         self.cg_function_space = fenics.FunctionSpace(self.mesh, "CG", 1)
         self.dg_function_space = fenics.FunctionSpace(self.mesh, "DG", 0)
 
-        self.mu_lame = fenics.Function(self.cg_function_space)
+        self.mu_lame: fenics.Function = fenics.Function(self.cg_function_space)
         self.mu_lame.vector().vec().set(1.0)
+        self.mu_lame.vector().apply("")
 
         if self.shape_scalar_product is None:
             # Use the default linear elasticity approach
@@ -399,31 +312,27 @@ class ShapeFormHandler(form_handler.FormHandler):
 
                 vol_max = self.volumes.vector().vec().max()[1]
                 self.volumes.vector().vec().scale(1 / vol_max)
+                self.volumes.vector().apply("")
 
             else:
                 self.volumes = fenics.Constant(1.0)
 
-            def eps(u):
+            def eps(u: fenics.Function) -> ufl.core.expr.Expr:
                 """Computes the symmetric gradient of a vector field ``u``.
 
-                Parameters
-                ----------
-                u : fenics.Function
-                    A vector field
+                Args:
+                    u: A vector field
 
-                Returns
-                -------
-                ufl.core.expr.Expr
+                Returns:
                     The symmetric gradient of ``u``
 
-
                 """
                 return fenics.Constant(0.5) * (fenics.grad(u) + fenics.grad(u).T)
 
             trial = fenics.TrialFunction(self.deformation_space)
             test = fenics.TestFunction(self.deformation_space)
 
-            self.riesz_scalar_product = (
+            self.riesz_scalar_product: ufl.Form = (
                 fenics.Constant(2)
                 * self.mu_lame
                 / self.volumes
@@ -454,7 +363,7 @@ class ShapeFormHandler(form_handler.FormHandler):
 
                 self.inhomogeneous_mu = True
 
-                self.options_mu = [
+                self.options_mu: List[List[Union[str, int, float]]] = [
                     ["ksp_type", "cg"],
                     ["pc_type", "hypre"],
                     ["pc_hypre_type", "boomeramg"],
@@ -555,9 +464,11 @@ class ShapeFormHandler(form_handler.FormHandler):
                         x.sqrtabs()
 
                     self.mu_lame.vector().vec().aypx(0.0, x)
+                    self.mu_lame.vector().apply("")
 
                 else:
                     self.mu_lame.vector().vec().set(self.mu_fix)
+                    self.mu_lame.vector().apply("")
 
             else:
                 self.distance.vector().vec().aypx(
@@ -568,15 +479,14 @@ class ShapeFormHandler(form_handler.FormHandler):
                     .vector()
                     .vec(),
                 )
+                self.distance.vector().apply("")
                 self.mu_lame.vector().vec().aypx(
                     0.0,
                     fenics.interpolate(self.mu_expression, self.cg_function_space)
                     .vector()
                     .vec(),
                 )
-
-            # for mpi compatibility
-            self.mu_lame.vector().apply("")
+                self.mu_lame.vector().apply("")
 
     def _project_scalar_product(self) -> None:
         """Ensures, that only free dimensions can be deformed."""
@@ -605,8 +515,10 @@ class ShapeFormHandler(form_handler.FormHandler):
                 .vector()
                 .vec(),
             )
+            self.volumes.vector().apply("")
             vol_max = self.volumes.vector().vec().max()[1]
             self.volumes.vector().vec().scale(1 / vol_max)
+            self.volumes.vector().apply("")
 
         self.assembler.assemble(self.fe_scalar_product_matrix)
         self.fe_scalar_product_matrix.ident_zeros()
@@ -626,6 +538,7 @@ class ShapeFormHandler(form_handler.FormHandler):
             The scalar product of a and b.
 
         """
+        result: float
         if (
             self.config.getboolean("ShapeGradient", "use_p_laplacian")
             and not self.uses_custom_scalar_product
@@ -660,7 +573,7 @@ class ShapeFormHandler(form_handler.FormHandler):
                 (p - 2) / 2.0,
             )
             # noinspection PyTypeChecker
-            self.p_laplace_form = (
+            self.p_laplace_form: ufl.Form = (
                 fenics.inner(
                     self.mu_lame
                     * (fenics.Constant(eps) + kappa)
