@@ -25,7 +25,6 @@ optimization problems.
 from __future__ import annotations
 
 import abc
-import configparser
 import copy
 import json
 from typing import Callable, Dict, List, Optional, TYPE_CHECKING, Union
@@ -38,14 +37,14 @@ from cashocs import _exceptions
 from cashocs import _loggers
 from cashocs import _utils
 from cashocs import io
+from cashocs._optimization import cost_functional
 
 if TYPE_CHECKING:
     from cashocs import _pde_problems
     from cashocs import types
-    from cashocs._optimization import cost_functional
-    from cashocs._optimization.optimization_variable_abstractions import (
-        OptimizationVariableAbstractions,
-    )
+    from cashocs._optimization import line_search as ls
+    from cashocs._optimization import optimization_algorithms
+    from cashocs._optimization import optimization_variable_abstractions as ova
 
 
 class OptimizationProblem(abc.ABC):
@@ -63,30 +62,41 @@ class OptimizationProblem(abc.ABC):
     gradient_problem: types.GradientProblem
     output_manager: io.OutputManager
     form_handler: types.FormHandler
-    optimization_variable_abstractions: OptimizationVariableAbstractions
+    optimization_variable_abstractions: ova.OptimizationVariableAbstractions
     adjoint_problem: _pde_problems.AdjointProblem
     state_problem: _pde_problems.StateProblem
     uses_custom_scalar_product: bool = False
+    use_scalar_tracking: bool = False
     temp_dict: Optional[Dict]
     algorithm: str
+    line_search: ls.LineSearch
+    hessian_problem: _pde_problems.HessianProblem
+    solver: optimization_algorithms.OptimizationAlgorithm
+    config: io.Config
+    initial_guess: Optional[List[fenics.Function]]
+    cost_functional_list: List[types.CostFunctional]
 
     def __init__(
         self,
-        state_forms: Union[ufl.Form, List[ufl.Form]],
+        state_forms: Union[List[ufl.Form], ufl.Form],
         bcs_list: Union[
-            fenics.DirichletBC, List[fenics.DirichletBC], List[List[fenics.DirichletBC]]
+            List[List[fenics.DirichletBC]], List[fenics.DirichletBC], fenics.DirichletBC
         ],
-        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
-        states: Union[fenics.Function, List[fenics.Function]],
-        adjoints: Union[fenics.Function, List[fenics.Function]],
+        cost_functional_form: Union[
+            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
+        ],
+        states: Union[List[fenics.Function], fenics.Function],
+        adjoints: Union[List[fenics.Function], fenics.Function],
         config: Optional[io.Config] = None,
-        initial_guess: Optional[List[fenics.Function]] = None,
-        ksp_options: Optional[Union[List[List[str]], List[List[List[str]]]]] = None,
-        adjoint_ksp_options: Optional[
-            Union[List[List[str]], List[List[List[str]]]]
+        initial_guess: Optional[Union[List[fenics.Function], fenics.Function]] = None,
+        ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
         ] = None,
-        scalar_tracking_forms: Optional[Dict] = None,
-        min_max_terms: Optional[Dict] = None,
+        adjoint_ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ] = None,
+        scalar_tracking_forms: Optional[Union[List[Dict], Dict]] = None,
+        min_max_terms: Optional[Union[List[Dict], Dict]] = None,
         desired_weights: Optional[List[float]] = None,
     ) -> None:
         r"""Initializes self.
@@ -149,13 +159,11 @@ class OptimizationProblem(abc.ABC):
         self.state_dim = len(self.state_forms)
         self.bcs_list = _utils.check_and_enlist_bcs(bcs_list)
 
-        self.cost_functional_list = _utils.enlist(cost_functional_form)
-        self.cost_functional_form = _utils.summation(self.cost_functional_list)
+        self._parse_cost_functional_form(cost_functional_form)
 
-        self.states = _utils.enlist(states)
-        self.adjoints = _utils.enlist(adjoints)
+        self.states: List[fenics.Function] = _utils.enlist(states)
+        self.adjoints: List[fenics.Function] = _utils.enlist(adjoints)
 
-        self.use_scalar_tracking = False
         self.use_min_max_terms = False
         self.use_scaling = False
 
@@ -171,10 +179,6 @@ class OptimizationProblem(abc.ABC):
 
         fenics.set_log_level(fenics.LogLevel.CRITICAL)
 
-        self.line_search = None
-        self.hessian_problem = None
-        self.solver = None
-
         self.has_custom_adjoint = False
         self.has_custom_derivative = False
 
@@ -209,15 +213,72 @@ class OptimizationProblem(abc.ABC):
         self.state_problem.has_solution = False
         self.adjoint_problem.has_solution = False
 
+    def _parse_cost_functional_form(
+        self,
+        cost_functional_form: Union[
+            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
+        ],
+    ) -> None:
+        """Parses the cost functional form for use in cashocs."""
+        self.input_cost_functional_list = _utils.enlist(cost_functional_form)
+        self.cost_functional_list = []
+        for functional in self.input_cost_functional_list:
+            if isinstance(functional, ufl.Form):
+                self.cost_functional_list.append(
+                    cost_functional.IntegralFunctional(functional)
+                )
+            elif isinstance(
+                functional,
+                (
+                    cost_functional.IntegralFunctional,
+                    cost_functional.ScalarTrackingFunctional,
+                    cost_functional.MinMaxFunctional,
+                ),
+            ):
+                self.cost_functional_list.append(functional)
+
+    def _parse_scalar_tracking_forms(
+        self, scalar_tracking_forms: Optional[Union[List[Dict], Dict]]
+    ) -> None:
+        self.scalar_tracking_legacy_mode = False
+        if scalar_tracking_forms is None:
+            self.scalar_tracking_forms = scalar_tracking_forms
+        else:
+            _loggers.warning(
+                "DEPRECATION WARNING: Using the keyword argument scalar_tracking_forms"
+                " is replaced by using cashocs.ScalarTrackingFunctional and appending"
+                "this to the list of cost functionals."
+                "The former usage will be removed in the future."
+            )
+            self.scalar_tracking_forms = None
+
+            for functional in _utils.enlist(scalar_tracking_forms):
+                integrand = functional["integrand"]
+                tracking_goal = functional["tracking_goal"]
+                try:
+                    weight = functional["weight"]
+                except KeyError:
+                    weight = 1.0
+                scalar_tracking_functional = cost_functional.ScalarTrackingFunctional(
+                    integrand, tracking_goal, weight
+                )
+                self.cost_functional_list.append(scalar_tracking_functional)
+            self.use_scalar_tracking = False
+            self.scalar_tracking_legacy_mode = True
+
     def _parse_optional_inputs(
         self,
-        config: configparser.ConfigParser,
-        initial_guess: List[fenics.Function],
-        ksp_options: Union[List[List[str]], List[List[List[str]]]],
-        adjoint_ksp_options: Union[List[List[str]], List[List[List[str]]]],
-        scalar_tracking_forms: Dict,
-        min_max_terms: Dict,
-        desired_weights: List[float],
+        config: Optional[io.Config],
+        initial_guess: Optional[Union[List[fenics.Function], fenics.Function]],
+        ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ],
+        adjoint_ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ],
+        scalar_tracking_forms: Optional[Union[List[Dict], Dict]],
+        min_max_terms: Optional[Union[List[Dict], Dict]],
+        desired_weights: Optional[Union[List[float], float]],
     ) -> None:
         """Initializes the optional input parameters.
 
@@ -255,8 +316,8 @@ class OptimizationProblem(abc.ABC):
             self.initial_guess = _utils.enlist(initial_guess)
 
         if ksp_options is None:
-            self.ksp_options = []
-            option = [
+            self.ksp_options: types.KspOptions = []
+            option: List[List[Union[str, int, float]]] = [
                 ["ksp_type", "preonly"],
                 ["pc_type", "lu"],
                 ["pc_factor_mat_solver_type", "mumps"],
@@ -268,23 +329,35 @@ class OptimizationProblem(abc.ABC):
         else:
             self.ksp_options = _utils.check_and_enlist_ksp_options(ksp_options)
 
-        self.adjoint_ksp_options = (
+        self.adjoint_ksp_options: types.KspOptions = (
             self.ksp_options[:]
             if adjoint_ksp_options is None
             else _utils.check_and_enlist_ksp_options(adjoint_ksp_options)
         )
 
-        if scalar_tracking_forms is None:
-            self.scalar_tracking_forms = scalar_tracking_forms
-        else:
-            self.scalar_tracking_forms = _utils.enlist(scalar_tracking_forms)
-            self.use_scalar_tracking = True
+        self._parse_scalar_tracking_forms(scalar_tracking_forms)
 
         if min_max_terms is None:
             self.min_max_terms = min_max_terms
         else:
-            self.use_min_max_terms = True
-            self.min_max_terms = _utils.enlist(min_max_terms)
+            _loggers.warning(
+                "DEPRECATION WARNING: Using the keyword argument min_max_terms"
+                " is replaced by using cashocs.MinMaxFunctional and appending"
+                "this to the list of cost functionals."
+                "The former usage will be removed in the future."
+            )
+            self.min_max_terms = None
+            self.use_min_max_terms = False
+            for functional in _utils.enlist(min_max_terms):
+                integrand = functional["integrand"]
+                lower_bound = functional["lower_bound"]
+                upper_bound = functional["upper_bound"]
+                mu = functional["mu"]
+                lambd = functional["lambda"]
+                min_max_functional = cost_functional.MinMaxFunctional(
+                    integrand, lower_bound, upper_bound, mu, lambd
+                )
+                self.cost_functional_list.append(min_max_functional)
 
         if desired_weights is None:
             self.desired_weights = desired_weights
@@ -292,16 +365,12 @@ class OptimizationProblem(abc.ABC):
             self.desired_weights = _utils.enlist(desired_weights)
             self.use_scaling = True
 
-            if scalar_tracking_forms is not None and not isinstance(
-                scalar_tracking_forms, list
-            ):
+            if self.use_scaling and self.scalar_tracking_legacy_mode:
                 raise _exceptions.InputError(
                     "OptimizationProblem",
                     "scalar_tracking_forms",
-                    (
-                        "If you supply desired weights, "
-                        "you have to supply a matching list in scalar_tracking_forms"
-                    ),
+                    "Scaling of scalar_tracking_forms is now only possible when using "
+                    "cashocs.ScalarTrackingFunctional",
                 )
 
     def compute_state_variables(self) -> None:
@@ -342,7 +411,7 @@ class OptimizationProblem(abc.ABC):
 
         """
         mod_forms = _utils.enlist(adjoint_forms)
-
+        mod_bcs_list: List
         if adjoint_bcs_list == [] or adjoint_bcs_list is None:
             mod_bcs_list = []
             for i in range(self.state_dim):
@@ -522,7 +591,7 @@ class OptimizationProblem(abc.ABC):
         self.form_handler.cost_functional_shift = shift
 
     @abc.abstractmethod
-    def gradient_test(self):
+    def gradient_test(self) -> float:
         """Test the correctness of the computed gradient with finite differences.
 
         Returns:
@@ -535,9 +604,9 @@ class OptimizationProblem(abc.ABC):
     def _compute_initial_function_values(self) -> None:
         """Computes the cost functional values for the initial iteration."""
         self.state_problem.solve()
-        self.initial_function_values = []
-        for i in range(len(self.cost_functional_list)):
-            val = fenics.assemble(self.cost_functional_list[i])
+        self.initial_function_values: List[float] = []
+        for i, functional in enumerate(self.cost_functional_list):
+            val = functional.evaluate()
 
             if abs(val) <= 1e-15:
                 val = 1.0
@@ -549,29 +618,7 @@ class OptimizationProblem(abc.ABC):
 
             self.initial_function_values.append(val)
 
-        if self.use_scalar_tracking:
-            self.initial_scalar_tracking_values = []
-            for i in range(len(self.scalar_tracking_forms)):
-                val = 0.5 * pow(
-                    fenics.assemble(
-                        self.form_handler.scalar_cost_functional_integrands[i]
-                    )
-                    - self.form_handler.scalar_tracking_goals[i],
-                    2,
-                )
-
-                if abs(val) <= 1e-15:
-                    val = 1.0
-                    _loggers.info(
-                        f"Term {i:d} of the scalar tracking cost functional "
-                        f"vanishes for the initial iteration. Multiplying "
-                        f"this term with the factor you supplied in desired "
-                        f"weights."
-                    )
-
-                self.initial_scalar_tracking_values.append(val)
-
-    def _scale_cost_functional(self):
+    def _scale_cost_functional(self) -> None:
         """Scales the terms of the cost functional and scalar_tracking forms."""
         _loggers.info(
             "You are using the automatic scaling functionality of cashocs."
@@ -579,27 +626,21 @@ class OptimizationProblem(abc.ABC):
             "functional yourself or if you supply custom forms."
         )
 
-        if not self.has_cashocs_remesh_flag:
-            self._compute_initial_function_values()
+        if self.use_scaling and self.desired_weights is not None:
+            if not self.has_cashocs_remesh_flag:
+                self._compute_initial_function_values()
 
-        else:
-            with open(f"{self.temp_dir}/temp_dict.json", "r", encoding="utf-8") as file:
-                temp_dict = json.load(file)
-            self.initial_function_values = temp_dict["initial_function_values"]
-            if self.use_scalar_tracking:
-                self.initial_scalar_tracking_values = temp_dict[
-                    "initial_scalar_tracking_values"
-                ]
+            else:
+                with open(
+                    f"{self.temp_dir}/temp_dict.json", "r", encoding="utf-8"
+                ) as file:
+                    temp_dict: Dict = json.load(file)
+                self.initial_function_values = temp_dict["initial_function_values"]
 
-        for i in range(len(self.cost_functional_list)):
-            const = fenics.Constant(
-                np.abs(self.desired_weights[i] / self.initial_function_values[i])
-            )
-            self.cost_functional_list[i] = const * self.cost_functional_list[i]
-
-        if self.use_scalar_tracking:
-            for i in range(len(self.scalar_tracking_forms)):
-                self.scalar_tracking_forms[-1 - i]["weight"] = abs(
-                    self.desired_weights[-1 - i]
-                    / self.initial_scalar_tracking_values[-1 - i]
+            for i, functional in enumerate(self.cost_functional_list):
+                scaling_factor = np.abs(
+                    self.desired_weights[i] / self.initial_function_values[i]
                 )
+                functional.scale(scaling_factor)
+                if isinstance(functional, cost_functional.IntegralFunctional):
+                    self.input_cost_functional_list[i] = functional.form
