@@ -19,18 +19,17 @@
 
 from __future__ import annotations
 
-import configparser
 import json
 import os
 import subprocess  # nosec B404
 import sys
 import tempfile
-from typing import Dict, List, Optional, Union
+from types import TracebackType
+from typing import Any, Dict, List, Optional, Type, TYPE_CHECKING, Union
 
 import dolfin.function.argument
 import fenics
 import numpy as np
-from typing_extensions import Literal
 import ufl
 import ufl.algorithms
 
@@ -46,6 +45,10 @@ from cashocs._optimization import optimization_problem
 from cashocs._optimization import verification
 from cashocs._optimization.shape_optimization import shape_variable_abstractions
 
+if TYPE_CHECKING:
+    from cashocs import io
+    from cashocs import types
+
 
 class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
     r"""A shape optimization problem.
@@ -60,26 +63,33 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
     on).
     """
 
+    temp_dict: Optional[Dict]
+
     def __new__(
         cls,
-        state_forms: Union[ufl.Form, List[ufl.Form]],
+        state_forms: Union[List[ufl.Form], ufl.Form],
         bcs_list: Union[
-            fenics.DirichletBC,
-            List[fenics.DirichletBC],
             List[List[fenics.DirichletBC]],
-            None,
+            List[fenics.DirichletBC],
+            fenics.DirichletBC,
         ],
-        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
-        states: Union[fenics.Function, List[fenics.Function]],
-        adjoints: Union[fenics.Function, List[fenics.Function]],
+        cost_functional_form: Union[
+            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
+        ],
+        states: Union[List[fenics.Function], fenics.Function],
+        adjoints: Union[List[fenics.Function], fenics.Function],
         boundaries: fenics.MeshFunction,
-        config: Optional[configparser.ConfigParser] = None,
+        config: Optional[io.Config] = None,
         shape_scalar_product: Optional[ufl.Form] = None,
         initial_guess: Optional[List[fenics.Function]] = None,
-        ksp_options: Optional[List[List[List[str]]]] = None,
-        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
-        scalar_tracking_forms: Optional[Dict] = None,
-        min_max_terms: Optional[Dict] = None,
+        ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ] = None,
+        adjoint_ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ] = None,
+        scalar_tracking_forms: Optional[Union[List[Dict], Dict]] = None,
+        min_max_terms: Optional[Union[List[Dict], Dict]] = None,
         desired_weights: Optional[List[float]] = None,
     ) -> ShapeOptimizationProblem:
         """Initializes self.
@@ -139,7 +149,7 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
 
         if use_scaling:
             unscaled_problem = super().__new__(cls)
-            unscaled_problem.__init__(
+            unscaled_problem.__init__(  # type: ignore
                 state_forms,
                 bcs_list,
                 cost_functional_form,
@@ -162,19 +172,16 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
                 problem.initial_function_values = (
                     unscaled_problem.initial_function_values
                 )
-                if unscaled_problem.use_scalar_tracking:
-                    problem.initial_scalar_tracking_values = (
-                        unscaled_problem.initial_scalar_tracking_values
-                    )
 
             if (
                 not unscaled_problem.has_cashocs_remesh_flag
                 and unscaled_problem.do_remesh
+                and fenics.MPI.rank(fenics.MPI.comm_world) == 0
             ):
-                subprocess.run(  # nosec B603
+                subprocess.run(  # nosec B603, B607
                     ["rm", "-r", unscaled_problem.temp_dir], check=True
                 )
-                subprocess.run(  # nosec B603
+                subprocess.run(  # nosec B603, B607
                     ["rm", "-r", unscaled_problem.mesh_handler.remesh_directory],
                     check=True,
                 )
@@ -186,24 +193,29 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
 
     def __init__(
         self,
-        state_forms: Union[ufl.Form, List[ufl.Form]],
+        state_forms: Union[List[ufl.Form], ufl.Form],
         bcs_list: Union[
-            fenics.DirichletBC,
-            List[fenics.DirichletBC],
             List[List[fenics.DirichletBC]],
-            None,
+            List[fenics.DirichletBC],
+            fenics.DirichletBC,
         ],
-        cost_functional_form: Union[ufl.Form, List[ufl.Form]],
-        states: Union[fenics.Function, List[fenics.Function]],
-        adjoints: Union[fenics.Function, List[fenics.Function]],
+        cost_functional_form: Union[
+            List[types.CostFunctional], types.CostFunctional, List[ufl.Form], ufl.Form
+        ],
+        states: Union[List[fenics.Function], fenics.Function],
+        adjoints: Union[List[fenics.Function], fenics.Function],
         boundaries: fenics.MeshFunction,
-        config: Optional[configparser.ConfigParser] = None,
+        config: Optional[io.Config] = None,
         shape_scalar_product: Optional[ufl.Form] = None,
         initial_guess: Optional[List[fenics.Function]] = None,
-        ksp_options: Optional[List[List[List[str]]]] = None,
-        adjoint_ksp_options: Optional[List[List[List[str]]]] = None,
-        scalar_tracking_forms: Optional[Dict] = None,
-        min_max_terms: Optional[Dict] = None,
+        ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ] = None,
+        adjoint_ksp_options: Optional[
+            Union[types.KspOptions, List[List[Union[str, int, float]]]]
+        ] = None,
+        scalar_tracking_forms: Optional[Union[List[Dict], Dict]] = None,
+        min_max_terms: Optional[Union[List[Dict], Dict]] = None,
         desired_weights: Optional[List[float]] = None,
     ) -> None:
         """Initializes self.
@@ -282,12 +294,11 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
 
         self.boundaries = boundaries
 
-        # shape_scalar_product
         self.shape_scalar_product = shape_scalar_product
         if shape_scalar_product is None:
-            self.deformation_space = None
+            self.deformation_space: Optional[fenics.FunctionSpace] = None
         else:
-            self.deformation_space = self.shape_scalar_product.arguments()[
+            self.deformation_space = shape_scalar_product.arguments()[
                 0
             ].ufl_function_space()
 
@@ -307,9 +318,13 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
             )
 
         self.is_shape_problem = True
-        self.form_handler = _forms.ShapeFormHandler(self)
+        self.form_handler: _forms.ShapeFormHandler = _forms.ShapeFormHandler(self)
 
-        if self.do_remesh and not self.has_cashocs_remesh_flag:
+        if (
+            self.do_remesh
+            and not self.has_cashocs_remesh_flag
+            and self.temp_dict is not None
+        ):
             # noinspection PyUnresolvedReferences
             self.temp_dict["Regularization"] = {
                 "mu_volume": self.form_handler.shape_regularization.mu_volume,
@@ -372,9 +387,14 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
 
             if not self.has_cashocs_remesh_flag:
                 self.directory = os.path.dirname(os.path.realpath(sys.argv[0]))
-                self.temp_dir = tempfile.mkdtemp(
-                    prefix="._cashocs_remesh_temp_", dir=self.directory
-                )
+                if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                    temp_dir: str = tempfile.mkdtemp(
+                        prefix="._cashocs_remesh_temp_", dir=self.directory
+                    )
+                else:
+                    temp_dir = ""
+                self.temp_dir: str = fenics.MPI.comm_world.bcast(temp_dir, root=0)
+
                 self._change_except_hook()
                 self.temp_dict = {
                     "temp_dir": self.temp_dir,
@@ -392,10 +412,6 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
                         self.temp_dict[
                             "initial_function_values"
                         ] = self.initial_function_values
-                        if self.use_scalar_tracking:
-                            self.temp_dict[
-                                "initial_scalar_tracking_values"
-                            ] = self.initial_scalar_tracking_values
                 except AttributeError:  # this happens for the unscaled problem
                     pass
 
@@ -419,17 +435,7 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
 
     def solve(
         self,
-        algorithm: Optional[
-            Literal[
-                "gradient_descent",
-                "gd",
-                "conjugate_gradient",
-                "nonlinear_cg",
-                "ncg",
-                "lbfgs",
-                "bfgs",
-            ]
-        ] = None,
+        algorithm: Optional[str] = None,
         rtol: Optional[float] = None,
         atol: Optional[float] = None,
         max_iter: Optional[int] = None,
@@ -515,7 +521,11 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
         (only needed for remeshing)
         """
 
-        def custom_except_hook(exctype, value, traceback) -> None:  # pragma: no cover
+        def custom_except_hook(
+            exctype: Type[BaseException],
+            value: BaseException,
+            traceback: TracebackType,
+        ) -> Any:  # pragma: no cover
             """A customized hook which is injected when an exception occurs.
 
             Args:
@@ -528,14 +538,20 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
                 "An exception was raised by cashocs, "
                 "deleting the created temporary files."
             )
-            if not self.config.getboolean("Debug", "remeshing"):
-                subprocess.run(["rm", "-r", self.temp_dir], check=True)  # nosec B603
-                subprocess.run(  # nosec B603
+            if (
+                not self.config.getboolean("Debug", "remeshing")
+                and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+            ):
+                subprocess.run(  # nosec B603, B607
+                    ["rm", "-r", self.temp_dir],
+                    check=True,
+                )
+                subprocess.run(  # nosec B603, B607
                     ["rm", "-r", self.mesh_handler.remesh_directory], check=True
                 )
             sys.__excepthook__(exctype, value, traceback)
 
-        sys.excepthook = custom_except_hook
+        sys.excepthook = custom_except_hook  # type: ignore
 
     def compute_shape_gradient(self) -> List[fenics.Function]:
         """Solves the Riesz problem to determine the shape gradient.
@@ -571,30 +587,11 @@ class ShapeOptimizationProblem(optimization_problem.OptimizationProblem):
                 {shape_derivative.arguments()[0]: self.form_handler.test_vector_field},
             )
 
-        retry_assembler_setup = False
-        if not self.form_handler.degree_estimation:
-            try:
-                self.form_handler.assembler = fenics.SystemAssembler(
-                    self.form_handler.riesz_scalar_product,
-                    shape_derivative,
-                    self.form_handler.bcs_shape,
-                )
-            except (AssertionError, ValueError):
-                retry_assembler_setup = True
-
-        if retry_assembler_setup or self.form_handler.degree_estimation:
-            estimated_degree = np.maximum(
-                ufl.algorithms.estimate_total_polynomial_degree(
-                    self.form_handler.riesz_scalar_product
-                ),
-                ufl.algorithms.estimate_total_polynomial_degree(shape_derivative),
-            )
-            self.form_handler.assembler = fenics.SystemAssembler(
-                self.form_handler.riesz_scalar_product,
-                shape_derivative,
-                self.form_handler.bcs_shape,
-                form_compiler_parameters={"quadrature_degree": estimated_degree},
-            )
+        self.form_handler.setup_assembler(
+            self.form_handler.riesz_scalar_product,
+            shape_derivative,
+            self.form_handler.bcs_shape,
+        )
 
         self.has_custom_derivative = True
 
