@@ -23,7 +23,7 @@ cost functional.
 
 from __future__ import annotations
 
-from typing import List, TYPE_CHECKING
+from typing import List, TYPE_CHECKING, Union
 
 import fenics
 from petsc4py import PETSc
@@ -67,11 +67,13 @@ class ControlGradientProblem(pde_problem.PDEProblem):
         # noinspection PyUnresolvedReferences
         self.ksps = [PETSc.KSP().create() for _ in range(self.form_handler.control_dim)]
 
-        gradient_tol = self.config.getfloat("OptimizationRoutine", "gradient_tol")
+        gradient_tol: float = self.config.getfloat(
+            "OptimizationRoutine", "gradient_tol"
+        )
 
-        gradient_method = self.config.get("OptimizationRoutine", "gradient_method")
+        gradient_method: str = self.config.get("OptimizationRoutine", "gradient_method")
 
-        option = []
+        option: List[List[Union[str, int, float]]] = []
         if gradient_method.casefold() == "direct":
             option = [
                 ["ksp_type", "preonly"],
@@ -114,9 +116,7 @@ class ControlGradientProblem(pde_problem.PDEProblem):
 
         if not self.has_solution:
             for i in range(self.form_handler.control_dim):
-                fenics.assemble(
-                    self.form_handler.gradient_forms_rhs[i], tensor=self.b_tensors[i]
-                )
+                self.form_handler.assemblers[i].assemble(self.b_tensors[i])
                 _utils.solve_linear_problem(
                     ksp=self.ksps[i],
                     b=self.b_tensors[i].vec(),
