@@ -23,7 +23,7 @@ in the truncated Newton method.
 
 from __future__ import annotations
 
-from typing import List, TYPE_CHECKING
+from typing import List, TYPE_CHECKING, Union
 
 import fenics
 import numpy as np
@@ -35,12 +35,36 @@ from cashocs import nonlinear_solvers
 
 if TYPE_CHECKING:
     from cashocs import _forms
+    from cashocs import types
     from cashocs._pde_problems import control_gradient_problem
 
 
 class HessianProblem:
     """PDE Problem used to solve the (reduced) Hessian problem."""
 
+    states_prime: List[fenics.Function]
+    adjoints_prime: List[fenics.Function]
+    bcs_list_ad: List[fenics.DirichletBC]
+    delta_control: List[fenics.Function]
+    residual: List[fenics.Function]
+    state_A_tensors: List[fenics.PETScMatrix]  # pylint: disable=invalid-name
+    adjoint_A_tensors: List[fenics.PETScMatrix]  # pylint: disable=invalid-name
+    state_b_tensors: List[fenics.PETScVector]
+    adjoint_b_tensors: List[fenics.PETScVector]
+    temp1: List[fenics.Function]
+    temp2: List[fenics.Function]
+    p: List[fenics.Function]
+    p_prev: List[fenics.Function]
+    p_pprev: List[fenics.Function]
+    s: List[fenics.Function]
+    s_prev: List[fenics.Function]
+    s_pprev: List[fenics.Function]
+    q: List[fenics.Function]
+    q_prev: List[fenics.Function]
+    hessian_actions: List[fenics.Function]
+    inactive_part: List[fenics.Function]
+    active_part: List[fenics.Function]
+
     def __init__(
         self,
         form_handler: _forms.ControlFormHandler,
@@ -66,88 +90,13 @@ class HessianProblem:
         self.inner_newton_atol = self.config.getfloat("AlgoTNM", "inner_newton_atol")
 
         self.test_directions = self.form_handler.test_directions
-        self.residual = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.delta_control = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
 
-        # pylint: disable=invalid-name
-        self.state_A_tensors = [
-            fenics.PETScMatrix() for _ in range(self.form_handler.state_dim)
-        ]
-        self.state_b_tensors = [
-            fenics.PETScVector() for _ in range(self.form_handler.state_dim)
-        ]
-        # pylint: disable=invalid-name
-        self.adjoint_A_tensors = [
-            fenics.PETScMatrix() for _ in range(self.form_handler.state_dim)
-        ]
-        self.adjoint_b_tensors = [
-            fenics.PETScVector() for _ in range(self.form_handler.state_dim)
-        ]
+        self._init_helper_functions()
+        self._init_linalg()
 
         self.state_dim = self.form_handler.state_dim
         self.control_dim = self.form_handler.control_dim
 
-        self.temp1 = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.temp2 = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-
-        self.p = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.p_prev = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.p_pprev = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.s = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.s_prev = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.s_pprev = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.q = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.q_prev = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-
-        self.hessian_actions = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.inactive_part = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-        self.active_part = [
-            fenics.Function(function_space)
-            for function_space in self.form_handler.control_spaces
-        ]
-
         self.controls = self.form_handler.controls
 
         self.picard_rtol = self.config.getfloat("StateSystem", "picard_rtol")
@@ -155,10 +104,6 @@ class HessianProblem:
         self.picard_max_iter = self.config.getint("StateSystem", "picard_iter")
         self.picard_verbose = self.config.getboolean("StateSystem", "picard_verbose")
 
-        self.states_prime = None
-        self.adjoints_prime = None
-        self.bcs_list_ad = None
-
         self.no_sensitivity_solves = 0
         # noinspection PyUnresolvedReferences
         self.state_ksps = [
@@ -177,7 +122,7 @@ class HessianProblem:
         # noinspection PyUnresolvedReferences
         self.ksps = [PETSc.KSP().create() for _ in range(self.control_dim)]
 
-        option = [
+        option: List[List[Union[str, int, float]]] = [
             ["ksp_type", "cg"],
             ["pc_type", "hypre"],
             ["pc_hypre_type", "boomeramg"],
@@ -186,7 +131,7 @@ class HessianProblem:
             ["ksp_atol", 1e-50],
             ["ksp_max_it", 100],
         ]
-        self.riesz_ksp_options = []
+        self.riesz_ksp_options: types.KspOptions = []
         for i in range(self.control_dim):
             self.riesz_ksp_options.append(option)
 
@@ -194,6 +139,47 @@ class HessianProblem:
         for i, ksp in enumerate(self.ksps):
             ksp.setOperators(self.form_handler.riesz_projection_matrices[i])
 
+    def _init_helper_functions(self) -> None:
+        """Initializes the helper functions."""
+        self.residual = _utils.create_function_list(self.form_handler.control_spaces)
+        self.delta_control = _utils.create_function_list(
+            self.form_handler.control_spaces
+        )
+        self.temp1 = _utils.create_function_list(self.form_handler.control_spaces)
+        self.temp2 = _utils.create_function_list(self.form_handler.control_spaces)
+        self.p = _utils.create_function_list(self.form_handler.control_spaces)
+        self.p_prev = _utils.create_function_list(self.form_handler.control_spaces)
+        self.p_pprev = _utils.create_function_list(self.form_handler.control_spaces)
+        self.s = _utils.create_function_list(self.form_handler.control_spaces)
+        self.s_prev = _utils.create_function_list(self.form_handler.control_spaces)
+        self.s_pprev = _utils.create_function_list(self.form_handler.control_spaces)
+        self.q = _utils.create_function_list(self.form_handler.control_spaces)
+        self.q_prev = _utils.create_function_list(self.form_handler.control_spaces)
+        self.hessian_actions = _utils.create_function_list(
+            self.form_handler.control_spaces
+        )
+        self.inactive_part = _utils.create_function_list(
+            self.form_handler.control_spaces
+        )
+        self.active_part = _utils.create_function_list(self.form_handler.control_spaces)
+
+    def _init_linalg(self) -> None:
+        """Initializes linear algebra matrices and vectors."""
+        # pylint: disable=invalid-name
+        self.state_A_tensors = [
+            fenics.PETScMatrix() for _ in range(self.form_handler.state_dim)
+        ]
+        self.state_b_tensors = [
+            fenics.PETScVector() for _ in range(self.form_handler.state_dim)
+        ]
+        # pylint: disable=invalid-name
+        self.adjoint_A_tensors = [
+            fenics.PETScMatrix() for _ in range(self.form_handler.state_dim)
+        ]
+        self.adjoint_b_tensors = [
+            fenics.PETScVector() for _ in range(self.form_handler.state_dim)
+        ]
+
     def hessian_application(
         self, h: List[fenics.Function], out: List[fenics.Function]
     ) -> None:
@@ -212,6 +198,7 @@ class HessianProblem:
         """
         for i in range(self.control_dim):
             self.test_directions[i].vector().vec().aypx(0.0, h[i].vector().vec())
+            self.test_directions[i].vector().apply("")
 
         self.states_prime = self.form_handler.states_prime
         self.adjoints_prime = self.form_handler.adjoints_prime
@@ -309,6 +296,7 @@ class HessianProblem:
         """
         for j in range(self.control_dim):
             out[j].vector().vec().set(0.0)
+            out[j].vector().apply("")
 
         self.form_handler.restrict_to_inactive_set(h, self.inactive_part)
         self.hessian_application(self.inactive_part, self.hessian_actions)
@@ -323,6 +311,7 @@ class HessianProblem:
                 self.active_part[j].vector().vec()
                 + self.inactive_part[j].vector().vec(),
             )
+            out[j].vector().apply("")
 
     def newton_solve(self) -> List[fenics.Function]:
         """Solves the Newton step with an iterative method.
@@ -336,6 +325,7 @@ class HessianProblem:
 
         for j in range(self.control_dim):
             self.delta_control[j].vector().vec().set(0.0)
+            self.delta_control[j].vector().apply("")
 
         if self.inner_newton.casefold() == "cg":
             self.cg()
@@ -348,7 +338,9 @@ class HessianProblem:
         """Solves the (truncated) Newton step with a CG method."""
         for j in range(self.control_dim):
             self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
+            self.residual[j].vector().apply("")
             self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
+            self.p[j].vector().apply("")
 
         rsold = self.form_handler.scalar_product(self.residual, self.residual)
         eps_0 = np.sqrt(rsold)
@@ -370,7 +362,9 @@ class HessianProblem:
                 self.delta_control[j].vector().vec().axpy(
                     alpha, self.p[j].vector().vec()
                 )
+                self.delta_control[j].vector().apply("")
                 self.residual[j].vector().vec().axpy(-alpha, self.q[j].vector().vec())
+                self.residual[j].vector().apply("")
 
             rsnew = self.form_handler.scalar_product(self.residual, self.residual)
             eps = np.sqrt(rsnew)
@@ -382,6 +376,7 @@ class HessianProblem:
 
             for j in range(self.control_dim):
                 self.p[j].vector().vec().aypx(beta, self.residual[j].vector().vec())
+                self.p[j].vector().apply("")
 
             rsold = rsnew
 
@@ -389,7 +384,9 @@ class HessianProblem:
         """Solves the (truncated) Newton step with a CR method."""
         for j in range(self.control_dim):
             self.residual[j].vector().vec().aypx(0.0, -self.gradient[j].vector().vec())
+            self.residual[j].vector().apply("")
             self.p[j].vector().vec().aypx(0.0, self.residual[j].vector().vec())
+            self.p[j].vector().apply("")
 
         eps_0 = np.sqrt(self.form_handler.scalar_product(self.residual, self.residual))
 
@@ -397,6 +394,7 @@ class HessianProblem:
 
         for j in range(self.control_dim):
             self.q[j].vector().vec().aypx(0.0, self.s[j].vector().vec())
+            self.q[j].vector().apply("")
 
         self.form_handler.restrict_to_active_set(self.residual, self.temp1)
         self.form_handler.restrict_to_active_set(self.s, self.temp2)
@@ -421,7 +419,9 @@ class HessianProblem:
                 self.delta_control[j].vector().vec().axpy(
                     alpha, self.p[j].vector().vec()
                 )
+                self.delta_control[j].vector().apply("")
                 self.residual[j].vector().vec().axpy(-alpha, self.q[j].vector().vec())
+                self.residual[j].vector().apply("")
 
             eps = np.sqrt(
                 self.form_handler.scalar_product(self.residual, self.residual)
@@ -447,6 +447,8 @@ class HessianProblem:
 
             for j in range(self.control_dim):
                 self.p[j].vector().vec().aypx(beta, self.residual[j].vector().vec())
+                self.p[j].vector().apply("")
                 self.q[j].vector().vec().aypx(beta, self.s[j].vector().vec())
+                self.q[j].vector().apply("")
 
             rar = rar_new
