@@ -25,7 +25,7 @@ import os
 import subprocess  # nosec B404
 import sys
 import tempfile
-from typing import List, TYPE_CHECKING
+from typing import cast, Dict, List, Optional, TYPE_CHECKING, Union
 
 import fenics
 import numpy as np
@@ -55,30 +55,30 @@ def _remove_gmsh_parametrizations(mesh_file: str) -> None:
 
     """
     temp_location = f"{mesh_file[:-4]}_temp.msh"
+    if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+        with open(mesh_file, "r", encoding="utf-8") as in_file, open(
+            temp_location, "w", encoding="utf-8"
+        ) as temp_file:
 
-    with open(mesh_file, "r", encoding="utf-8") as in_file, open(
-        temp_location, "w", encoding="utf-8"
-    ) as temp_file:
+            parametrizations_section = False
 
-        parametrizations_section = False
+            for line in in_file:
 
-        for line in in_file:
+                if line == "$Parametrizations\n":
+                    parametrizations_section = True
 
-            if line == "$Parametrizations\n":
-                parametrizations_section = True
+                if not parametrizations_section:
+                    temp_file.write(line)
+                else:
+                    pass
 
-            if not parametrizations_section:
-                temp_file.write(line)
-            else:
-                pass
+                if line == "$EndParametrizations\n":
+                    parametrizations_section = False
 
-            if line == "$EndParametrizations\n":
-                parametrizations_section = False
-
-    subprocess.run(["mv", temp_location, mesh_file], check=True)  # nosec B603
+        subprocess.run(["mv", temp_location, mesh_file], check=True)  # nosec B603, B607
 
 
-def filter_sys_argv(temp_dir: str):  # pragma: no cover
+def filter_sys_argv(temp_dir: str) -> List[str]:  # pragma: no cover
     """Filters the command line arguments for the cashocs remesh flag.
 
     Args:
@@ -110,6 +110,10 @@ class _MeshHandler:
     transformations and remeshing. Also includes mesh quality control checks.
     """
 
+    current_mesh_quality: float
+    mesh_quality_measure: str
+    temp_dict: Optional[Dict]
+
     def __init__(self, shape_optimization_problem: ShapeOptimizationProblem) -> None:
         """Initializes self.
 
@@ -129,8 +133,12 @@ class _MeshHandler:
         self.volume_change = float(self.config.get("MeshQuality", "volume_change"))
         self.angle_change = float(self.config.get("MeshQuality", "angle_change"))
 
-        self.mesh_quality_tol_lower = self.config.getfloat("MeshQuality", "tol_lower")
-        self.mesh_quality_tol_upper = self.config.getfloat("MeshQuality", "tol_upper")
+        self.mesh_quality_tol_lower: float = self.config.getfloat(
+            "MeshQuality", "tol_lower"
+        )
+        self.mesh_quality_tol_upper: float = self.config.getfloat(
+            "MeshQuality", "tol_upper"
+        )
 
         if self.mesh_quality_tol_lower > 0.9 * self.mesh_quality_tol_upper:
             _loggers.warning(
@@ -151,22 +159,28 @@ class _MeshHandler:
         self._setup_a_priori()
 
         # Remeshing initializations
-        self.do_remesh = self.config.getboolean("Mesh", "remesh")
-        self.save_optimized_mesh = self.config.getboolean("Output", "save_mesh")
+        self.do_remesh: bool = self.config.getboolean("Mesh", "remesh")
+        self.save_optimized_mesh: bool = self.config.getboolean("Output", "save_mesh")
 
         if self.do_remesh or self.save_optimized_mesh:
             self.mesh_directory = os.path.dirname(
                 os.path.realpath(self.config.get("Mesh", "gmsh_file"))
             )
 
-        if self.do_remesh:
+        if self.do_remesh and shape_optimization_problem.temp_dict is not None:
             self.temp_dict = shape_optimization_problem.temp_dict
-            self.gmsh_file = self.temp_dict["gmsh_file"]
+            self.gmsh_file: str = self.temp_dict["gmsh_file"]
             self.remesh_counter = self.temp_dict.get("remesh_counter", 0)
 
             if not self.form_handler.has_cashocs_remesh_flag:
-                self.remesh_directory = tempfile.mkdtemp(
-                    prefix="cashocs_remesh_", dir=self.mesh_directory
+                if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                    remesh_directory: str = tempfile.mkdtemp(
+                        prefix="cashocs_remesh_", dir=self.mesh_directory
+                    )
+                else:
+                    remesh_directory = ""
+                self.remesh_directory: str = fenics.MPI.comm_world.bcast(
+                    remesh_directory, root=0
                 )
             else:
                 self.remesh_directory = self.temp_dict["remesh_directory"]
@@ -182,9 +196,10 @@ class _MeshHandler:
             self.gmsh_file_init = (
                 f"{self.remesh_directory}/mesh_{self.remesh_counter:d}.msh"
             )
-            subprocess.run(  # nosec 603
-                ["cp", self.gmsh_file, self.gmsh_file_init], check=True
-            )
+            if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                subprocess.run(  # nosec 603
+                    ["cp", self.gmsh_file, self.gmsh_file_init], check=True
+                )
             self.gmsh_file = self.gmsh_file_init
 
     def move_mesh(self, transformation: fenics.Function) -> bool:
@@ -230,7 +245,7 @@ class _MeshHandler:
 
     def _setup_decrease_computation(self) -> None:
         """Initializes attributes and solver for the frobenius norm check."""
-        self.options_frobenius = [
+        self.options_frobenius: List[List[Union[str, int, float]]] = [
             ["ksp_type", "preonly"],
             ["pc_type", "jacobi"],
             ["pc_jacobi_type", "diagonal"],
@@ -291,6 +306,7 @@ class _MeshHandler:
             self.search_direction_container.vector().vec().aypx(
                 0.0, search_direction[0].vector().vec()
             )
+            self.search_direction_container.vector().apply("")
             x = _utils.assemble_and_solve_linear(
                 self.a_frobenius,
                 self.l_frobenius,
@@ -298,7 +314,7 @@ class _MeshHandler:
                 ksp_options=self.options_frobenius,
             )
 
-            frobenius_norm = np.max(x[:])
+            frobenius_norm = x.max()[1]
             beta_armijo = self.config.getfloat("OptimizationRoutine", "beta_armijo")
 
             return int(
@@ -313,7 +329,7 @@ class _MeshHandler:
 
     def _setup_a_priori(self) -> None:
         """Sets up the attributes and petsc solver for the a priori quality check."""
-        self.options_prior = [
+        self.options_prior: List[List[Union[str, int, float]]] = [
             ["ksp_type", "preonly"],
             ["pc_type", "jacobi"],
             ["pc_jacobi_type", "diagonal"],
@@ -360,6 +376,7 @@ class _MeshHandler:
         self.transformation_container.vector().vec().aypx(
             0.0, transformation.vector().vec()
         )
+        self.transformation_container.vector().apply("")
         x = _utils.assemble_and_solve_linear(
             self.A_prior,
             self.l_prior,
@@ -367,10 +384,12 @@ class _MeshHandler:
             ksp_options=self.options_prior,
         )
 
-        min_det = np.min(x[:])
-        max_det = np.max(x[:])
+        min_det: float = x.min()[1]
+        max_det: float = x.max()[1]
 
-        return (min_det >= 1 / self.volume_change) and (max_det <= self.volume_change)
+        return bool(
+            (min_det >= 1 / self.volume_change) and (max_det <= self.volume_change)
+        )
 
     def _generate_remesh_geo(self, input_mesh_file: str) -> None:
         """Generates a .geo file used for remeshing.
@@ -383,69 +402,89 @@ class _MeshHandler:
                 file.
 
         """
-        with open(self.remesh_geo_file, "w", encoding="utf-8") as file:
-            temp_name = os.path.split(input_mesh_file)[1]
+        if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+            with open(self.remesh_geo_file, "w", encoding="utf-8") as file:
+                temp_name = os.path.split(input_mesh_file)[1]
 
-            file.write(f"Merge '{temp_name}';\n")
-            file.write("CreateGeometry;\n")
-            file.write("\n")
+                file.write(f"Merge '{temp_name}';\n")
+                file.write("CreateGeometry;\n")
+                file.write("\n")
 
-            geo_file = self.temp_dict["geo_file"]
-            with open(geo_file, "r", encoding="utf-8") as f:
-                for line in f:
-                    if line[0].islower():
-                        file.write(line)
-                    if line[:5] == "Field":
-                        file.write(line)
-                    if line[:16] == "Background Field":
-                        file.write(line)
-                    if line[:19] == "BoundaryLayer Field":
-                        file.write(line)
-                    if line[:5] == "Mesh.":
-                        file.write(line)
+                self.temp_dict = cast(Dict, self.temp_dict)
+                geo_file = self.temp_dict["geo_file"]
+                with open(geo_file, "r", encoding="utf-8") as f:
+                    for line in f:
+                        if line[0].islower():
+                            file.write(line)
+                        if line[:5] == "Field":
+                            file.write(line)
+                        if line[:16] == "Background Field":
+                            file.write(line)
+                        if line[:19] == "BoundaryLayer Field":
+                            file.write(line)
+                        if line[:5] == "Mesh.":
+                            file.write(line)
 
     def clean_previous_gmsh_files(self) -> None:
         """Removes the gmsh files from the previous remeshing iterations."""
         gmsh_file = f"{self.remesh_directory}/mesh_{self.remesh_counter - 1:d}.msh"
-        if os.path.isfile(gmsh_file):
+        if os.path.isfile(gmsh_file) and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
             subprocess.run(["rm", gmsh_file], check=True)  # nosec 603
 
         gmsh_pre_remesh_file = (
             f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}_pre_remesh.msh"
         )
-        if os.path.isfile(gmsh_pre_remesh_file):
+        if (
+            os.path.isfile(gmsh_pre_remesh_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", gmsh_pre_remesh_file], check=True)  # nosec 603
 
         mesh_h5_file = f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}.h5"
-        if os.path.isfile(mesh_h5_file):
+        if os.path.isfile(mesh_h5_file) and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
             subprocess.run(["rm", mesh_h5_file], check=True)  # nosec 603
 
         mesh_xdmf_file = f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}.xdmf"
-        if os.path.isfile(mesh_xdmf_file):
+        if (
+            os.path.isfile(mesh_xdmf_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", mesh_xdmf_file], check=True)  # nosec 603
 
         boundaries_h5_file = (
             f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}_boundaries.h5"
         )
-        if os.path.isfile(boundaries_h5_file):
+        if (
+            os.path.isfile(boundaries_h5_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", boundaries_h5_file], check=True)  # nosec 603
 
         boundaries_xdmf_file = (
             f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}_boundaries.xdmf"
         )
-        if os.path.isfile(boundaries_xdmf_file):
+        if (
+            os.path.isfile(boundaries_xdmf_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", boundaries_xdmf_file], check=True)  # nosec 603
 
         subdomains_h5_file = (
             f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}_subdomains.h5"
         )
-        if os.path.isfile(subdomains_h5_file):
+        if (
+            os.path.isfile(subdomains_h5_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", subdomains_h5_file], check=True)  # nosec 603
 
         subdomains_xdmf_file = (
             f"{self.remesh_directory}/mesh_{self.remesh_counter-1:d}_subdomains.xdmf"
         )
-        if os.path.isfile(subdomains_xdmf_file):
+        if (
+            os.path.isfile(subdomains_xdmf_file)
+            and fenics.MPI.rank(fenics.MPI.comm_world) == 0
+        ):
             subprocess.run(["rm", subdomains_xdmf_file], check=True)  # nosec 603
 
     def _restart_script(self, temp_dir: str) -> None:
@@ -456,6 +495,7 @@ class _MeshHandler:
 
         """
         if not self.config.getboolean("Debug", "restart"):
+            sys.stdout.flush()
             os.execv(  # nosec 606
                 sys.executable,
                 [sys.executable]
@@ -470,7 +510,7 @@ class _MeshHandler:
                 "Restart of script with remeshed geometry is cancelled."
             )
 
-    def remesh(self, solver: OptimizationAlgorithm):
+    def remesh(self, solver: OptimizationAlgorithm) -> None:
         """Remeshes the current geometry with Gmsh.
 
         Performs a remeshing of the geometry, and then restarts the optimization problem
@@ -480,7 +520,7 @@ class _MeshHandler:
             solver: The optimization algorithm used to solve the problem.
 
         """
-        if self.do_remesh:
+        if self.do_remesh and self.temp_dict is not None:
             self.remesh_counter += 1
             temp_file = (
                 f"{self.remesh_directory}/mesh_{self.remesh_counter:d}_pre_remesh.msh"
@@ -519,14 +559,15 @@ class _MeshHandler:
                 "-o",
                 new_gmsh_file,
             ]
-            if not self.config.getboolean("Mesh", "show_gmsh_output"):
-                subprocess.run(  # nosec 603
-                    gmsh_cmd_list,
-                    check=True,
-                    stdout=subprocess.DEVNULL,
-                )
-            else:
-                subprocess.run(gmsh_cmd_list, check=True)  # nosec 603
+            if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                if not self.config.getboolean("Mesh", "show_gmsh_output"):
+                    subprocess.run(  # nosec 603
+                        gmsh_cmd_list,
+                        check=True,
+                        stdout=subprocess.DEVNULL,
+                    )
+                else:
+                    subprocess.run(gmsh_cmd_list, check=True)  # nosec 603
 
             _remove_gmsh_parametrizations(new_gmsh_file)
 
@@ -536,10 +577,11 @@ class _MeshHandler:
 
             new_xdmf_file = f"{self.remesh_directory}/mesh_{self.remesh_counter:d}.xdmf"
 
-            subprocess.run(  # nosec 603
-                ["cashocs-convert", new_gmsh_file, new_xdmf_file],
-                check=True,
-            )
+            if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                subprocess.run(  # nosec 603
+                    ["cashocs-convert", new_gmsh_file, new_xdmf_file],
+                    check=True,
+                )
 
             self.clean_previous_gmsh_files()
 
@@ -555,7 +597,8 @@ class _MeshHandler:
 
             temp_dir = self.temp_dict["temp_dir"]
 
-            with open(f"{temp_dir}/temp_dict.json", "w", encoding="utf-8") as file:
-                json.dump(self.temp_dict, file)
+            if fenics.MPI.rank(fenics.MPI.comm_world) == 0:
+                with open(f"{temp_dir}/temp_dict.json", "w", encoding="utf-8") as file:
+                    json.dump(self.temp_dict, file)
 
             self._restart_script(temp_dir)
