@@ -19,10 +19,11 @@
 
 from __future__ import annotations
 
+from typing import List, Union
+
 import fenics
 import numpy as np
 from petsc4py import PETSc
-from typing_extensions import Literal
 import ufl
 
 from cashocs import _utils
@@ -31,10 +32,8 @@ from cashocs.geometry import measure
 
 def compute_mesh_quality(
     mesh: fenics.Mesh,
-    quality_type: Literal["min", "minimum", "avg", "average"] = "min",
-    quality_measure: Literal[
-        "skewness", "maximum_angle", "radius_ratios", "condition_number"
-    ] = "skewness",
+    quality_type: str = "min",
+    quality_measure: str = "skewness",
 ) -> float:
     """This computes the mesh quality of a given mesh.
 
@@ -305,7 +304,8 @@ PYBIND11_MODULE(SIGNATURE, m)
             The minimum skewness of the mesh.
 
         """
-        return np.min(cls._quality_object.skewness(mesh).array())
+        quality: float = np.min(cls._quality_object.skewness(mesh).array())
+        return quality
 
     @classmethod
     def avg_skewness(cls, mesh: fenics.Mesh) -> float:
@@ -336,7 +336,8 @@ PYBIND11_MODULE(SIGNATURE, m)
             The average skewness of the mesh.
 
         """
-        return np.average(cls._quality_object.skewness(mesh).array())
+        quality: float = np.average(cls._quality_object.skewness(mesh).array())
+        return quality
 
     @classmethod
     def min_maximum_angle(cls, mesh: fenics.Mesh) -> float:
@@ -361,7 +362,8 @@ PYBIND11_MODULE(SIGNATURE, m)
             The minimum value of the maximum angle quality measure.
 
         """
-        return np.min(cls._quality_object.maximum_angle(mesh).array())
+        quality: float = np.min(cls._quality_object.maximum_angle(mesh).array())
+        return quality
 
     @classmethod
     def avg_maximum_angle(cls, mesh: fenics.Mesh) -> float:
@@ -386,7 +388,8 @@ PYBIND11_MODULE(SIGNATURE, m)
             The average quality, based on the maximum angle measure.
 
         """
-        return np.average(cls._quality_object.maximum_angle(mesh).array())
+        quality: float = np.average(cls._quality_object.maximum_angle(mesh).array())
+        return quality
 
     @classmethod
     def min_radius_ratios(cls, mesh: fenics.Mesh) -> float:
@@ -409,7 +412,8 @@ PYBIND11_MODULE(SIGNATURE, m)
 
         """
         # noinspection PyArgumentList
-        return np.min(fenics.MeshQuality.radius_ratios(mesh).array())
+        quality: float = np.min(fenics.MeshQuality.radius_ratios(mesh).array())
+        return quality
 
     @classmethod
     def avg_radius_ratios(cls, mesh: fenics.Mesh) -> float:
@@ -432,7 +436,8 @@ PYBIND11_MODULE(SIGNATURE, m)
 
         """
         # noinspection PyArgumentList
-        return np.average(fenics.MeshQuality.radius_ratios(mesh).array())
+        quality: float = np.average(fenics.MeshQuality.radius_ratios(mesh).array())
+        return quality
 
     @classmethod
     def min_condition_number(cls, mesh: fenics.Mesh) -> float:
@@ -453,7 +458,7 @@ PYBIND11_MODULE(SIGNATURE, m)
         jac = ufl.Jacobian(mesh)
         inv = ufl.JacobianInverse(mesh)
 
-        options = [
+        options: List[List[Union[str, int, float]]] = [
             ["ksp_type", "preonly"],
             ["pc_type", "jacobi"],
             ["pc_jacobi_type", "diagonal"],
@@ -485,9 +490,12 @@ PYBIND11_MODULE(SIGNATURE, m)
         )
         cond.vector().apply("")
         cond.vector().vec().reciprocal()
+        cond.vector().apply("")
         cond.vector().vec().scale(np.sqrt(mesh.geometric_dimension()))
+        cond.vector().apply("")
 
-        return cond.vector().vec().min()[1]
+        quality: float = cond.vector().vec().min()[1]
+        return quality
 
     @classmethod
     def avg_condition_number(cls, mesh: fenics.Mesh) -> float:
@@ -508,7 +516,7 @@ PYBIND11_MODULE(SIGNATURE, m)
         jac = ufl.Jacobian(mesh)
         inv = ufl.JacobianInverse(mesh)
 
-        options = [
+        options: List[List[Union[str, int, float]]] = [
             ["ksp_type", "preonly"],
             ["pc_type", "jacobi"],
             ["pc_jacobi_type", "diagonal"],
@@ -541,6 +549,9 @@ PYBIND11_MODULE(SIGNATURE, m)
         cond.vector().apply("")
 
         cond.vector().vec().reciprocal()
+        cond.vector().apply("")
         cond.vector().vec().scale(np.sqrt(mesh.geometric_dimension()))
+        cond.vector().apply("")
 
-        return cond.vector().vec().sum() / cond.vector().vec().getSize()
+        quality: float = cond.vector().vec().sum() / cond.vector().vec().getSize()
+        return quality
