@@ -35,6 +35,11 @@ from cashocs import _utils
 class _NewtonSolver:
     """A Newton solver."""
 
+    assembler_shift: fenics.SystemAssembler
+    residual_shift: fenics.PETScVector
+    A_matrix: fenics.PETScMatrix  # pylint: disable=invalid-name
+    b: fenics.PETScVector
+
     # noinspection PyUnresolvedReferences,PyPep8Naming
     def __init__(
         self,
@@ -52,7 +57,7 @@ class _NewtonSolver:
         inexact: bool = True,
         verbose: bool = True,
         ksp: Optional[PETSc.KSP] = None,
-        ksp_options: Optional[List[List[str]]] = None,
+        ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
         A_tensor: Optional[fenics.PETScMatrix] = None,  # pylint: disable=invalid-name
         b_tensor: Optional[fenics.PETScVector] = None,
         is_linear: bool = False,
@@ -111,12 +116,16 @@ class _NewtonSolver:
         self.norm_type = norm_type
         self.damped = damped
         self.inexact = inexact
-        self.verbose = verbose
         self.A_tensor = A_tensor  # pylint: disable=invalid-name
         self.b_tensor = b_tensor
         self.is_linear = is_linear
 
-        self.derivative = derivative or fenics.derivative(self.nonlinear_form, self.u)
+        self.verbose = verbose if not self.is_linear else False
+
+        temp_derivative = derivative or fenics.derivative(self.nonlinear_form, self.u)
+        self.derivative = _utils.bilinear_boundary_form_modification([temp_derivative])[
+            0
+        ]
 
         # Setup increment and function for monotonicity test
         self.function_space = u.function_space()
@@ -165,17 +174,12 @@ class _NewtonSolver:
         self.A_fenics = self.A_tensor or fenics.PETScMatrix()
         self.residual = self.b_tensor or fenics.PETScVector()
 
-        self.assembler_shift = None
-        self.residual_shift = None
-
         if self.shift is not None:
             self.assembler_shift = fenics.SystemAssembler(
                 self.derivative, self.shift, self.bcs_hom
             )
             self.residual_shift = fenics.PETScVector()
 
-        self.b = None
-        self.A_matrix = None  # pylint: disable=invalid-name
         self.breakdown = False
         self.res = 1.0
         self.res_0 = 1.0
@@ -183,18 +187,21 @@ class _NewtonSolver:
 
     def _print_output(self) -> None:
         """Prints the output of the current iteration to the console."""
-        if self.verbose:
+        if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
             print(
                 f"Newton Iteration {self.iterations:2d} - "
                 f"residual (abs):  {self.res:.3e} (tol = {self.atol:.3e})    "
-                f"residual (rel):  {self.res / self.res_0:.3e} (tol = {self.rtol:.3e})"
+                f"residual (rel):  {self.res / self.res_0:.3e} "
+                f"(tol = {self.rtol:.3e})"
             )
 
     def _assemble_matrix(self) -> None:
         """Assembles the matrix for solving the linear problem."""
         self.assembler.assemble(self.A_fenics)
         self.A_fenics.ident_zeros()
-        self.A_matrix = fenics.as_backend_type(self.A_fenics).mat()
+        self.A_matrix = fenics.as_backend_type(  # pylint: disable=invalid-name
+            self.A_fenics
+        ).mat()
 
     def _compute_eta_inexact(self) -> None:
         """Computes the parameter ``eta`` for the inexact Newton method."""
@@ -219,6 +226,11 @@ class _NewtonSolver:
         if self.is_linear:
             self.eta = self.rtol * 1e-1
 
+    def _check_for_nan_residual(self) -> None:
+        """Checks, whether the residual is nan. If yes, raise a NotConvergedError."""
+        if np.isnan(self.res):
+            raise _exceptions.NotConvergedError("newton solver", "Residual is nan.")
+
     def solve(self) -> fenics.Function:
         r"""Solves the (nonlinear) problem with Newton\'s method.
 
@@ -230,11 +242,12 @@ class _NewtonSolver:
 
         self.res_0 = self.residual.norm(self.norm_type)
         if self.res_0 == 0.0:  # pragma: no cover
-            if self.verbose:
+            if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
                 print("Residual vanishes, input is already a solution.")
             return self.u
 
         self.res = self.res_0
+        self._check_for_nan_residual()
         self._print_output()
 
         self.tol = self._compute_convergence_tolerance()
@@ -247,6 +260,7 @@ class _NewtonSolver:
             self.lmbd = 1.0
             self.breakdown = False
             self.u_save.vector().vec().aypx(0.0, self.u.vector().vec())
+            self.u_save.vector().apply("")
 
             self._compute_eta_inexact()
             _utils.solve_linear_problem(
@@ -261,12 +275,14 @@ class _NewtonSolver:
 
             if self.is_linear:
                 self.u.vector().vec().axpy(1.0, self.du.vector().vec())
+                self.u.vector().apply("")
                 break
 
             if self.damped:
                 self._backtracking_line_search()
             else:
                 self.u.vector().vec().axpy(1.0, self.du.vector().vec())
+                self.u.vector().apply("")
 
             self._compute_residual()
 
@@ -275,17 +291,25 @@ class _NewtonSolver:
 
             res_prev = self.res
             self.res = self.residual.norm(self.norm_type)
+            self._check_for_nan_residual()
             self.eta_a = self.gamma * pow(self.res / res_prev, 2)
             self._print_output()
 
             if self.res <= self.tol:
-                if self.verbose:
+                if self.verbose and fenics.MPI.rank(fenics.MPI.comm_world) == 0:
                     print(
                         f"\nNewton Solver converged "
                         f"after {self.iterations:d} iterations.\n"
                     )
                 break
 
+        if self.res > self.tol and not self.is_linear:
+            raise _exceptions.NotConvergedError(
+                "Newton solver",
+                f"The Newton solver did not converge after "
+                f"{self.iterations:d} iterations.",
+            )
+
         return self.u
 
     def _check_for_divergence(self) -> None:
@@ -328,6 +352,7 @@ class _NewtonSolver:
         """Performs a backtracking line search for the damped Newton method."""
         while True:
             self.u.vector().vec().axpy(self.lmbd, self.du.vector().vec())
+            self.u.vector().apply("")
             self._compute_residual()
             _utils.solve_linear_problem(
                 ksp=self.ksp,
@@ -346,6 +371,7 @@ class _NewtonSolver:
                 break
             else:
                 self.u.vector().vec().aypx(0.0, self.u_save.vector().vec())
+                self.u.vector().apply("")
                 self.lmbd /= 2
 
             if self.lmbd < 1e-6:
@@ -369,7 +395,7 @@ def newton_solve(
     inexact: bool = True,
     verbose: bool = True,
     ksp: Optional[PETSc.KSP] = None,
-    ksp_options: Optional[List[List[str]]] = None,
+    ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
     A_tensor: Optional[fenics.PETScMatrix] = None,  # pylint: disable=invalid-name
     b_tensor: Optional[fenics.PETScVector] = None,
     is_linear: bool = False,
@@ -478,7 +504,7 @@ def damped_newton_solve(
     damped: bool = True,
     verbose: bool = True,
     ksp: Optional[PETSc.KSP] = None,
-    ksp_options: Optional[List[List[str]]] = None,
+    ksp_options: Optional[List[List[Union[str, int, float]]]] = None,
 ) -> fenics.Function:  # pragma: no cover
     r"""Damped Newton solve interface, only here for compatibility reasons.
 
